/*
 * Copyright (C) 2005, 2007-2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "basetools.h"

#include <iostream>
#include <stdlib.h>
#include <cstring>

#include "../config.h"
#include "../shared/platform.h"
#include "../shared/lib_conversions.h"
#include "../shared/hirestime.h"
#include "../shared/readjpeg.h"
#include "../shared/imgxform.h"

#ifdef USE_EPEG
# include <Epeg.h>
#endif

#ifdef USE_INTEGRATED_EPEG
# include "../lib_epeg/Epeg.h"
#endif

#ifdef USE_LIBGRAPHICSMAGICK
# include <magick/api.h>
#endif

#include "../shared/lib_conversions.h"
#include "../shared/filetools.h"
#include "../shared/thread.h"
#include "workerdata.h"

#define CREATETHUMBSTEST 0

void* createThumbs(void* wd)
{
  int nRetCode = BaseTools::createThumbs((WorkerData*)wd);
  
  SimpleThread::Exit();  
  return((void*)nRetCode);
}

void* copyFiles(void* wd)
{
  int nRetCode = BaseTools::copyFiles((WorkerData*)wd);
    
  SimpleThread::Exit();
  return((void*)nRetCode);
}

void BaseTools::getShrinkSize(int w, int h, int& wmax, int& hmax)
{
  if (w > wmax || h > hmax)
  {
    // An der Dimension orientieren, die eher am Limit ist
    if ((float)w/wmax > (float)h/hmax)
      // Die Breite ist limitierend
      hmax = (int)(0.5 + wmax * ((float)h/w));
    else
      // Die Höhe ist limitierend
      wmax  = (int)(0.5 + hmax * ((float)w/h));
  }
  else
  {
    // no need to shrink
    wmax = w;
    hmax = h;
  }
}

int BaseTools::createThumb(std::string strDst, std::string strSrc,
                           float quality, int nSpeed, 
                           int wmax, int hmax)
{
  bool fDelSrc = false;
  bool fDone = false;
  std::string tmpfilepath;
  int tempfd = TemporaryTextFile::GetTempFD(tmpfilepath);      
  if (tempfd != -1)
    close(tempfd);

  std::cout << "thumb from " << strSrc << "\n";
  std::cout << "  " << nSpeed << " ";
  
  bool fJPEG;
  {
    std::string suff = strU(FileTools::getSuffix(strSrc));
    fJPEG = suff.compare(".JPEG") == 0 || suff.compare(".JPG") == 0;
  }
  
#if (USE_EPEG == 1 || USE_INTEGRATED_EPEG == 1)
  // If the source file is a jpeg, I can use epeg to speed things up.
  if (nSpeed == 2 && fJPEG)
  {
    std::cout << "epeg ";
    
    Epeg_Image *im;
    int w, h;
    
    im = epeg_file_open(strSrc.c_str());
    if (!im)
    {
      std::cerr << "Cannot open " << strSrc << "\n";
      return(-1);
    }
    
    epeg_size_get(im, &w, &h);
    // std::cout << "epeg: src is " << w << "x" << h;
      
    if (w > wmax || h > hmax)
    {
      double fw = ((double)(w))/wmax;
      double fh = ((double)(h))/hmax;
      
      if (fw > fh)
      {
        h = (h*wmax)/w;
        w = wmax;
      }
      else
      {
        w = (w*hmax)/h;
        h = hmax;
      }
      
      // std::cout << ", strDst is " << w << "x" << h << ", qual=" << quality << "\n";
      epeg_decode_size_set(im, w, h);
      epeg_quality_set(im, (int)(quality*100));
      epeg_thumbnail_comments_enable(im, 0);
      epeg_file_output_set(im, strDst.c_str());
      epeg_encode(im);        
      fDone = true;
    }
    epeg_close(im);
  }
#endif
  
  if (!fDone && nSpeed < 2 && fJPEG)
  {
    std::cout << "internal ";

    // 
    int nW = wmax;
    int nH = hmax;
    unsigned char* array = LoadScaledJPEG(strSrc.c_str(), nW, nH);
    if (array)
    {
      int nD = 3;
      
      getShrinkSize(nW, nH, wmax, hmax);
      if (nW >= wmax*2 && nH >= hmax*2)
      {
        float flDiv  = (1.0*nW)/wmax;
        float flDivH = (1.0*nH)/hmax;
        if (flDiv > flDivH)
          flDiv = flDivH;
              
        unsigned char* tmp = array;
        array = ImgXForm::Shrink(tmp, nW, nH, nD, floor(flDiv));
        delete[] tmp;
      }
      
      unsigned char* array_scaled = ImgXForm::Bilinear(array, nW, nH, nD, wmax, hmax);
      delete[] array;
      
      if (WriteJPEG(strDst.c_str(), array_scaled, wmax, hmax, nD, quality*100) == 0)
        fDone = true;
    }
  }
  
  if (!fDone)
  {
#ifdef USE_LIBGRAPHICSMAGICK
    std::cout << "libgraphicsmagick ";
    
    ExceptionInfo exception;
    Image* image;
    Image* resize_image;
    ImageInfo* image_info;
    
    GetExceptionInfo(&exception);
    image_info=CloneImageInfo((ImageInfo *) NULL);
    
    strcpy(image_info->filename, strSrc.c_str());    
    image=ReadImage(image_info, &exception);
    CatchException(&exception);
    if (image != 0)
    {
      getShrinkSize(image->columns, image->rows, wmax, hmax);
      // siehe http://www.graphicsmagick.org/api/resize.html
      //resize_image=ResizeImage(image, wmax, hmax, LanczosFilter, 1.0, &exception); // slow
      if (nSpeed > 1)
        resize_image = SampleImage(image, wmax, hmax, &exception); // fast, nearest neighbor
      else        
        resize_image = ScaleImage(image, wmax, hmax, &exception); // medium
    
      CatchException(&exception);
      if (resize_image)
      {
        strcpy(resize_image->filename, strDst.c_str());
        image_info->quality = (int)(quality*100 + 0.5);
        if (WriteImage(image_info, resize_image) == 0)
        {
          CatchException(&(resize_image->exception));
          std::cerr << "error1\n";
          return(-1);
        }
      }
      else
      {
        std::cerr << "error2\n";
        return(-1);
      }
      
      DestroyImage(image);
      DestroyImage(resize_image);
      DestroyImageInfo(image_info);
      DestroyExceptionInfo(&exception);
      
      fDone = true;
    }
#endif
  }
    
  if (!fDone)
  {
    std::cout << "convert ";
    
    std::string sizestring = itoStr(wmax, ' ', 1) + "x" + itoStr(hmax, ' ', 1);
    std::string qualstring = ftoStr(quality, 1, 2);
    
    // jwtodo: Option, um den Algorithmus zum resize zwischen schnell und hochqualitativ
    // umzuschalten. Möglicherweise geht das mit '-filter'.
    //
    // Anstelle von '-thumbnail' kann man auch '-resize' benutzen, aber das braucht hier
    // doppelt so lange...
    std::string call = "";
#ifndef WIN32
    call = "nice -19 ";
#endif
    call += "convert \""
      + strSrc
      + "\" -flatten -thumbnail " + sizestring
      + " -quality " + qualstring
      + " \"" + strDst + "\"";
    
    if (system(call.c_str()))
    {
      std::cerr << call << "\n";
      return(-1);
    }
  }
  
  if (fDelSrc)
  {
    FileTools::rm(strSrc);
  }
  
  std::cout << "\n";
  
  return(0);
}

int BaseTools::createThumbs(WorkerData* wd)
{
  int               nRetCode = 0;
  WorkThumbCreate*  work     = (WorkThumbCreate*)((WorkThumbCreate*)wd->work);
  unsigned int      uSize    = work->filelist.size();

  HiResTime hrt;
  hrt.Capture();
  
  std::cout << "--- Creating thumbnails ------------------------- \n";
  
  for (unsigned int uCnt=0; (uCnt<uSize) && (wd->nStop == 0); uCnt+= 2)
  {
    if (FileTools::makeSureDirExists(work->filelist[uCnt+1].substr(0, work->filelist[uCnt+1].rfind('/'))))
      nRetCode = -1;
    
    if (createThumb(work->filelist[uCnt+1], work->filelist[uCnt], 
                    work->flQual, work->nType,
                    work->nMaxW, work->nMaxH))
      nRetCode = -1;
    
    //
    wd->uNumUnitsDone++;
  }

  std::cout << "timing.thumbnails=" << hrt.GetMSecSinceCapture() << "\n";
  
  ((WorkerData*)wd)->nRun     = 0;
  ((WorkerData*)wd)->nRetCode = nRetCode;
  
  return(nRetCode);
}

int BaseTools::copyFiles(WorkerData* wd)
{
  int                  nRetCode = 0;
  WorkThumbCreate*     work     = (WorkThumbCreate*)((WorkThumbCreate*)wd->work);
  unsigned int         uSize    = work->filelist.size();
  
  std::cout << "--- Copying files ------------------------------- \n";
  
  for (unsigned int uCnt=0; (uCnt<uSize) && (wd->nStop == 0); uCnt+= 2)
  {
    if (FileTools::makeSureDirExists(work->filelist[uCnt+1].substr(0, work->filelist[uCnt+1].rfind('/'))))
      nRetCode = -1;
    
    std::cout << "copying " << work->filelist[uCnt] << " to " << work->filelist[uCnt+1] << "\n";
    
#if CREATETHUMBSTEST == 0
    if (FileTools::copy(work->filelist[uCnt+1], work->filelist[uCnt]))
      nRetCode = -1;
#endif
    
    //
    wd->uNumUnitsDone++;
  }
  
  ((WorkerData*)wd)->nRun     = 0;
  ((WorkerData*)wd)->nRetCode = nRetCode;
  
  return(nRetCode);
}

std::string BaseTools::Keyword(std::string in)
{
  in = trim(collapse_space(in));
  for (std::string::size_type n=0; n<in.length(); n++)
  {
    if (in[n] == '.' || in[n] == ',')
      in[n] = '_';
  }
  return(in);
}

std::string BaseTools::KeywordList_String(std::string in)
{
  std::string::size_type pos, start;
  
  for (pos=0; pos<in.length(); pos++)
  {
    if (in[pos] == '\n')
      in[pos] = ',';
    else if (isspace(in[pos]))
      in[pos] = ' ';
  }
  
  in = trim(collapse_space(collapse_char(in, ',')));
  
  start = 0;
  while ((pos = in.find(", ", start)) != std::string::npos)
  {
    in = in.replace(pos, 2, ",");
    start = pos;
  }
  
  start = 0;
  while ((pos = in.find(" ,", start)) != std::string::npos)
  {
    in = in.replace(pos, 2, ",");
    start = pos;
  }
  
  in = collapse_char(in, ',');
  
  pos = in.length();   
  if (pos > 0 && in[pos-1] == ',')
    in = in.substr(0, pos-1);
  if (in.length() > 0 && in[0] == ',')
    in = in.substr(1);
      
  return(in);
}

std::set<std::string> BaseTools::KeywordList_Set(std::string in)
{
  std::string::size_type pos, start;
  std::set<std::string>  kwset;
  std::string            kw;
  
  for (pos=0; pos<in.length(); pos++)
  {
    if (in[pos] == '\n')
      in[pos] = ',';
    else if (isspace(in[pos]))
      in[pos] = ' ';
  }
  
  in = trim(collapse_space(collapse_char(in, ',')));
  
  pos   = in.find(',');
  start = 0;
  
  while (pos != std::string::npos)
  {
    if (pos > start)
    {
      kw = trim(in.substr(start, pos-start));
      if (kw.length())
        kwset.insert(kw);
    }
    start = pos+1;
    pos   = in.find(',', start);
  }
  if (start == 0 && in.length()) // only one keyword, without trailing ','
  {
    kw = trim(in);
    if (kw.length())
      kwset.insert(kw);
  }
  else if (start < in.length()) // the last keyword, without trailing ','
  {
    kw = trim(in.substr(start, in.length()-start));
    if (kw.length())
      kwset.insert(kw);
  }
  
  return(kwset);
}

std::string BaseTools::getFilePath(std::string relname)
{
  const char* installpaths[] = { "../",                     // win32
                                 STR_INSTALL_PREFIX "/",    // unix
                                 "/usr/local/"         ,    // unix, fallback
                                 "/opt/"               ,    // unix, fallback
                                 "/usr/"                };  // unix, fallback
  
  std::string fpath = FileTools::getHomeDir("Fotoalbum") + relname;
  if (FileTools::fileExists(fpath))
    return(fpath);
  
  for (unsigned int n=0; n<sizeof(installpaths)/sizeof(const char*); n++)
  {
    fpath = std::string(installpaths[n]) + relname;
    std::cout << "trying " << fpath << "\n";
    if (FileTools::fileExists(fpath))
      return(fpath);
  }
 
  return("");
}
