// -*- mode: c; mode: fold -*-
//
/*
 * Copyright (C) 2006-2008, 2010-2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "export.h"

#include <iostream>
#include "treeitem.h"
#include "treegroupbase.h"
#include "treepicbase.h"
#include "treeview.h"
#include "strings.h"
#include "exif.h"
#include "basetools.h"
#include "../shared/filetools.h"
#include "../shared/lib_conversions.h"
#include "../../version.h"

#if (EXIV2_FOUND == 1)
# include <exiv2/image.hpp>
# include <exiv2/exif.hpp>
# include <exiv2/iptc.hpp>
#endif

void Export::setOpts(SimpleXMLTransfer* exopts)
{
  SimpleXMLTransfer* xExp = exopts->getChild("export");
  
  type        = (eType)(xExp->attributeAsInt("type"));
  structure   = (eStructure)(xExp->attributeAsInt("structure"));
  fRecursive  = (xExp->attributeAsInt("recurse")     != 0);
  fThumbs     = (xExp->attributeAsInt("thumbs")      != 0);
  fFollowLinks= (xExp->attributeAsInt("followlinks") != 0);
  fDatafile   = (xExp->attributeAsInt("datafile")    != 0);
  fAddToJpeg  = (xExp->attributeAsInt("addtojpeg")   != 0);
  fPrefix     = (xExp->attributeAsInt("prefix")      != 0);
  fHtml       = (xExp->attributeAsInt("html")        != 0);
  uThumbSize  =  xExp->attributeAsInt("thumbsize", 200);
  uThumbQual  =  xExp->attributeAsDouble("thumbqual", 0.75)*100;
  uShrinkMode =  xExp->attributeAsInt("fast", 0);
  {
    SimpleXMLTransfer* xShrink = exopts->getChild("shrink");
    uShrinkQual =  xShrink->attributeAsDouble("quality", 0.75)*100;
    xShrink = xShrink->getChildAt(xShrink->attributeAsInt("index"));
    uShrinkW = xShrink->attributeAsInt("w");
    uShrinkH = xShrink->attributeAsInt("h");
  }
  
  nPicsInARow = 760 / (uThumbSize+3);
  
  // --- Constraints ------------------------------------------------
  if (fHtml)
  {
    structure = eHTML;
    fThumbs   = true;
    fPrefix   = false;
  }
}

std::string Export::getUniqueName(SetOfStrings&  namesInDir,
                                  std::string    name)
{
  if (namesInDir.find(name) == namesInDir.end())
  {
    namesInDir.insert(name);
    return(name);
  }
  else
  {
    std::string  tmp;
    unsigned int cnt=0;
    unsigned int cntcopy;

    while (cnt<1048576)
    {
      tmp = "";
      cntcopy = cnt;
      for (unsigned int i=0; i<5; i++)
      {
        tmp += (char)((cntcopy&0xF) + 'a');
        cntcopy >>= 4;
      }
      tmp += name;

      if (namesInDir.find(tmp) == namesInDir.end())
      {
        namesInDir.insert(tmp);
        return(tmp);
      }

      cnt++;
    }

    // Da finde ich wohl keinen einzigartigen Namen.
    if (error.length() == 0)
      error = "Unable to create unique name. Lots of files with the same name in one directory?";
    return(name);
  }
}

std::string Export::GetHtmlName(int nNr)
{
  if (nNr >= 0)
    return(itoStr(nNr, '0', 6) + ".htm");
  else
    return("index.htm");
}

void Export::HtmlLinks(std::string   link_up,
                       int           nNr,
                       bool          fNext,
                       std::string&  strLinkPrev,
                       std::string&  strLinkNext,
                       std::string&  strLinkUp,
                       std::string&  strURLNext)
{
  strLinkPrev = "<span class=\"arr\"><a href=\"" + GetHtmlName(nNr-1) + "\" class=\"arr\">&larr;</a></span>";
  
  if (fNext)
  {
    strURLNext = GetHtmlName(nNr+1);
    strLinkNext = "<span class=\"arr\"><a href=\"" + strURLNext + "\" class=\"arr\">&rarr;</a></span>";
  }
  else
  {
    strURLNext = ""; // todo
    strLinkNext = "<span class=\"ari\">&rarr;</span>";
  }
  
  if (link_up.empty())      
    strLinkUp = "<span class=\"ari\">&uarr;</span>";
  else
    strLinkUp = "<span class=\"arr\"><a href=\"" + link_up + "\" class=\"arr\">&uarr;</a></span>";
}

void Export::WriteConvExif(std::ostream& out, std::string in)
{
  out << "<p>";

  for (std::string::size_type i=0; i<in.length(); )
  {
    if (in[i] == '\n')
    {
      out << "<br/>";
      i++;
    }
    else
    {
      out << in[i];
      i++;
    }
  }

  out << "</p>";
}

void Export::picture(TreePicBase*       item,
                     std::string        pathOfNames,
                     SetOfStrings&      namesInDir,
                     SimpleXMLTransfer* meta,
                     int                nNr,
                     std::string&       groupout,
                     int&               nType,
                     bool               fNext,
                     std::string        strUp)
{
  std::string  somepath;
  std::string  dstpath;
  unsigned int cnt;
  std::string  name_out;
  std::string  name_html = "";
  std::string  html_thumbpath;

  TreeItem::eItemLinkType linktype = item->getLinkType();

  // --- Create metadata --------------------------------------------
  SimpleXMLTransfer* mymeta = new SimpleXMLTransfer();

  if (linktype == TreeItem::DATA)
  {
    std::string tmp;

    mymeta->setName("itemdata");
    if (item->hasDate())
      mymeta->setAttribute("date",     item->getDateString());
    if (item->hasTime())
      mymeta->setAttribute("time",     item->getTimeString());

    tmp = item->getKeywords();
    if (tmp.length())
      mymeta->setAttribute("keywords", tmp);

    tmp = item->getComment();
    if (tmp.length())
      mymeta->setAttribute("comment",  tmp);
  }
  else
  {
    mymeta->setName("itemlink");
    mymeta->setAttribute("ref",     item->getLinkID());
  }

  if (item->getID())
    mymeta->setAttribute("id",       item->getID());
  
  bool fCreateDst = true;
  if (fFollowLinks || item->getLinkType() == TreeItem::DATA)
  {
    // --- Care about the file itself ---------------------------------
    somepath = data->getPicPath(item);
    std::string name_in = FileTools::name(somepath);

    switch (structure)
    {
      case eFlat:
        name_out = getUniqueName(namesInDir, name_in);
        dstpath  = name_out;
        break;

      case eFileSys:
        dstpath  = item->getText();
        name_out = name_in;
        break;
        
      case eHTML:
        name_out = pCache->UpToDate(somepath, destdir, uShrinkW, uShrinkH, uShrinkMode, uShrinkQual);
        if (name_out.empty())
          name_out = pCache->Insert(somepath, destdir, uShrinkW, uShrinkH, uShrinkMode, uShrinkQual);
        else
          fCreateDst = false;
        name_html = itoStr(nNr, '0', 6) + ".htm";
        dstpath = name_out;
        break;
        
      default:
        if (fPrefix)
          name_in = itoStr(nNr, '0', 3) + "_" + name_in;
        name_out = getUniqueName(namesInDir, name_in);
        dstpath  = pathOfNames + name_out;
        break;
    }

    if (linktype == TreeItem::DATA)
      mymeta->setAttribute("path", dstpath);

    dstpath = destdir + dstpath;
    
    if (fCreateDst)
    {
      switch (type)
      {
        case eCopy:
          copy.push_back(somepath);
          copy.push_back(dstpath);
          break;
          
        case eShrink:
          shrink.push_back(somepath);
          shrink.push_back(dstpath);
          break;
          
        default:
          break;
      }
      annotate.push_back(PathAndPtr(dstpath, item));
    }
    
    // --- Maybe there are thumbnails to copy? ------------------------
    if (fThumbs && !fHtml)
    {
      std::string sizestr;

      cnt = 0;
      while (TreeView::uThumbSizes[cnt] > 0)
      {
        somepath = data->getPicPath(item, TreeView::uThumbSizes[cnt]);

        if (FileTools::fileExists(somepath))
        {
          copy.push_back(somepath);

          somepath = FileTools::path(dstpath) + ".thumbnails/";
          sizestr  = itoStr(TreeView::uThumbSizes[cnt], ' ', 0);
          somepath += sizestr + "x" + sizestr + "/" + name_out;

          copy.push_back(somepath);
        }

        cnt++;
      }
    }

    if (fHtml)
    {
      html_thumbpath = pCache->UpToDate(somepath, destdir, uThumbSize, uThumbSize, uShrinkMode, uThumbQual);
      if (html_thumbpath.empty())
      {
        html_thumbpath = pCache->Insert(somepath, destdir, uThumbSize, uThumbSize, uShrinkMode, uThumbQual);
        vThumbs.push_back(somepath);
        vThumbs.push_back(destdir + html_thumbpath);
      }
    }
  }
  meta->addChild(mymeta);

  // --- html export ------------------------------------------------
  if (fHtml)
  {
    groupout += "  <td align=\"center\"><a href=\"" + name_html + "\"><img src=\"" 
             + html_thumbpath + "\" alt=\"thumb\" border=\"0\"/></a></td>\n";

    if (++nType == (nPicsInARow)+1)
    {
      groupout += "</tr>\n<tr>\n";
      nType = 1;
    }

    std::string strLinkPrev, strLinkNext, strLinkUp, strURLNext;
    std::ofstream out;
    FileTools::makeSureDirExists(destdir);
    out.open((destdir + name_html).c_str());
    
    HtmlLinks(strUp, nNr, fNext, strLinkPrev, strLinkNext, strLinkUp, strURLNext);

    for (unsigned int n=0; n<templ_pic->list.size(); n++)
    {
      switch (templ_pic->list[n].nID)
      {
        case eTVURLNext:
          out << strURLNext;
          break;          
        case eTVHead:
          out << templ_head;
          break;
        case eTVButPrev:
          out << strLinkPrev;
          break;
        case eTVButNext:
          out << strLinkNext;
          break;
        case eTVButUp:
          out << strLinkUp;
          break;
        case eTVDateTime:
          out << item->getDateString() << " " << item->getTimeString();
          break;
        case eTVFilename:
          out << name_out;
          break;
        case eTVPath:
          out << item->getText();
          break;
        case eTVExif:
          {
            ExifInfo* exif = new ExifInfo(data->getPicPath(item));
            if (exif->hasEXIF())
            {
              WriteConvExif(out, exif->getInfo());
            }
          }
          break;
        case eTVComment:
          out << item->getCommentHtml();
          break;
        case eTVKeywords:
          out << item->getKeywords();
          break;

        case eTVFA_URL:
          out << FA_URL;
          break;

        case eTVFA_Name:
          out << FA_NAME;
          break;

        case eTVFA_Version:
          out << FOTOALBUMVERSION;
          break;

        default:
          out << templ_pic->list[n].value;
          break;
      }
    }
    out.close();
  }
}

void Export::groupToFileHtml(std::string path,
                             std::string strLinkPrev,
                             std::string strLinkNext,
                             std::string strLinkUp,
                             int         nPrevGroup,
                             int         nNextGroup,
                             std::string strText,
                             std::string strComment,
                             std::string strLinks,
                             std::string strThumbs,
                             std::string strKeywords)
{
  std::ofstream htmlfile;

  FileTools::makeSureDirExists(FileTools::path(path));
  htmlfile.open(path.c_str());

  for (unsigned int n=0; n<templ_group->list.size(); n++)
  {
    switch (templ_group->list[n].nID)
    {
      case eTVHead:
        htmlfile << templ_head;
        break;
      case eTVButPrev:
        htmlfile << strLinkPrev;
        break;
      case eTVButNext:
        htmlfile << strLinkNext;
        break;
      case eTVButUp:
        htmlfile << strLinkUp;
        break;
      case eTVButNextGroup:
        if (nNextGroup>=0)
          htmlfile << "<span class=\"arr\"><a href=\"" + GetHtmlName(nNextGroup) + "\" class=\"arr\">&raquo;</a></span>";
        else
          htmlfile << "<span class=\"ari\">&raquo;</span>";
        break;
      case eTVButPrevGroup:
        if (nPrevGroup>=0)
          htmlfile << "<span class=\"arr\"><a href=\"" + GetHtmlName(nPrevGroup) + "\" class=\"arr\">&laquo;</a></span>";
        else
          htmlfile << "<span class=\"ari\">&laquo;</span>";
        break;
      case eTVTitle:
        htmlfile << strText;
        break;
      case eTVComment:
        htmlfile << strComment;
        break;
      case eTVLinks:
        htmlfile << strLinks;
        break;
      case eTVThumbs:
        if (strThumbs.length())
          htmlfile << "<table><tr>" << strThumbs << "</tr></table>";
        break;
      case eTVKeywords:
        htmlfile << strKeywords;
        break;

      case eTVFA_URL:
        htmlfile << FA_URL;
        break;

      case eTVFA_Name:
        htmlfile << FA_NAME;
        break;

      case eTVFA_Version:
        htmlfile << FOTOALBUMVERSION;
        break;

      default:
        htmlfile << templ_group->list[n].value;
        break;
    }
  }

  htmlfile.close();
}

void Export::group(TreeGroupBase*     item,
                   std::string        pathOfNames,
                   SetOfStrings&      namesInDir,
                   SimpleXMLTransfer* meta,
                   int&               nNr,
                   std::string&       groupout,
                   int&               nType,
                   bool               fNext,
                   std::string        strUp,
                   int&               nPrevGroup)
{
  std::string      name;
  SetOfStrings     namesInMe;
  SetOfStrings*    namesptr;
  int              htmltype = 1;
  int              nMyPrevGroup = nPrevGroup;

  TreeItem::eItemLinkType linktype = item->getLinkType();

  SimpleXMLTransfer* mymeta = new SimpleXMLTransfer();

  switch (structure)
  {
    case eView:
      if (fHtml)
        name = itoStr(nNr, '0', 6);
      else
      {
        name = FileTools::cleanString(item->getText());
        if (fPrefix)
          name = itoStr(nNr, '0', 3) + "_" + name;
        name = getUniqueName(namesInDir, name);
      }
      pathOfNames += name + "/";
      namesptr = &namesInMe;
      break;

    case eFlat:
      namesptr = &namesInDir;
      break;

    default:
      namesptr = &namesInMe;
      break;
  }

  // --- Create metadata --------------------------------------------
  if (linktype != TreeItem::LINKPROPERTIES)
  {
    std::string tmp;

    mymeta->setName("group");
    tmp = item->getText();
    if (tmp.length())
      mymeta->setAttribute("name", tmp);

    tmp = item->getKeywords();
    if (tmp.length())
      mymeta->setAttribute("keywords", tmp);

    tmp = item->getComment();
    if (tmp.length())
      mymeta->setAttribute("comment",  tmp);
  }
  else
  {
    mymeta->setName("grouplinkproperties");
    mymeta->setAttribute("ref", item->getLinkID());
  }

  if (item->getID())
    mymeta->setAttribute("id", item->getID());

  // export all my children -- separate for "linearized browsing" first
  std::vector<TreePicBase*>   listPics;
  std::vector<TreeGroupBase*> listGroups;
  TreeItem* i = item->itemFirstChild;
  while (i != (TreeItem*)0)
  {
    if (i->getType() == TreeItem::PIC)
      listPics.push_back((TreePicBase*)(i));
    else if (fRecursive)
      listGroups.push_back((TreeGroupBase*)(i));
    i = i->itemNext;
  }
  
  int nMyNum = nNr;
  std::string strFile = GetHtmlName(nNr++);
  
  std::string strThumbtable = "";
  for (unsigned int n=0; n<listPics.size(); n++)
  {
    TreePicBase* i = listPics[n];
    picture(i, 
            pathOfNames, *namesptr, mymeta, nNr++, strThumbtable, htmltype, 
            fNext || !listGroups.empty() || (n<listPics.size()-1),
            strFile);
  }
  
  int nFirstGroupInsideOfMe = -1;
  if (!listGroups.empty())
    nFirstGroupInsideOfMe = nNr;
  nPrevGroup = nMyNum;
  std::string strLinks = "";
  for (unsigned int n=0; n<listGroups.size(); n++)
  {
    TreeGroupBase* i = listGroups[n];    
    group(i, 
          pathOfNames, *namesptr, mymeta, nNr, strLinks, htmltype,
          fNext || (n<listGroups.size()-1),
          strFile, nPrevGroup);
  }

  //
  meta->addChild(mymeta);

  // --- html export ------------------------------------------------
  if (fHtml)
  {
    groupout += "<a href=\"" + strFile + "\">" + item->getText() + "</a><br/>\n";
    nType = 0;

    std::string strLinkPrev, strLinkNext, strLinkUp, strURLNext;
    HtmlLinks(strUp, nMyNum, 
              fNext || !listPics.empty() || !listGroups.empty(),
              strLinkPrev, strLinkNext, strLinkUp, strURLNext);
    
    int nNextGroup = -1;
    if (nFirstGroupInsideOfMe >= 0)
      nNextGroup = nFirstGroupInsideOfMe;
    else if (fNext)
      nNextGroup = nNr;
    groupToFileHtml(destdir + strFile,
                    strLinkPrev, strLinkNext, strLinkUp,
                    nMyPrevGroup, nNextGroup,
                    item->getText(), item->getCommentHtml(),
                    strLinks, strThumbtable, item->getKeywords());
  }
}

void Export::PrepareTemplate(Template* templ)
{
  for (unsigned int n=0; n<templ->list.size(); n++)
  {
    // std::cout << templ->list[n].nID << " " << templ->list[n].value << "\n----------------------------------------\n\n";
    if (templ->list[n].nID == 0)
    {
      std::string val = templ->list[n].value;
      if (val.compare("{HEAD}") == 0)
        templ->list[n].nID = eTVHead;
      else if (val.compare("{BUTTON_PREV}") == 0)
        templ->list[n].nID = eTVButPrev;
      else if (val.compare("{BUTTON_NEXT}") == 0)
        templ->list[n].nID = eTVButNext;
      else if (val.compare("{BUTTON_PREV_GROUP}") == 0)
        templ->list[n].nID = eTVButPrevGroup;
      else if (val.compare("{BUTTON_NEXT_GROUP}") == 0)
        templ->list[n].nID = eTVButNextGroup;
      else if (val.compare("{BUTTON_UP}") == 0)
        templ->list[n].nID = eTVButUp;
      else if (val.compare("{URL_NEXT}") == 0)
        templ->list[n].nID = eTVURLNext;
//      else if (val.compare("{BUTTON_LIN_NEXT}") == 0)
//        templ->list[n].nID = eTVButLinNext;
      else if (val.compare("{DATETIME}") == 0)
        templ->list[n].nID = eTVDateTime;
      else if (val.compare("{FILENAME}") == 0)
        templ->list[n].nID = eTVFilename;
      else if (val.compare("{PATH}") == 0)
        templ->list[n].nID = eTVPath;
      else if (val.compare("{EXIF}") == 0)
        templ->list[n].nID = eTVExif;
      else if (val.compare("{COMMENT}") == 0)
        templ->list[n].nID = eTVComment;
      else if (val.compare("{KEYWORDS}") == 0)
        templ->list[n].nID = eTVKeywords;
      else if (val.compare("{TITLE}") == 0)
        templ->list[n].nID = eTVTitle;
      else if (val.compare("{LINKS}") == 0)
        templ->list[n].nID = eTVLinks;
      else if (val.compare("{THUMBTABLE}") == 0)
        templ->list[n].nID = eTVThumbs;
      else if (val.compare("{FA_URL}") == 0)
        templ->list[n].nID = eTVFA_URL;
      else if (val.compare("{FA_Name}") == 0)
        templ->list[n].nID = eTVFA_Name;
      else if (val.compare("{FA_Version}") == 0)
        templ->list[n].nID = eTVFA_Version;
      else
      {
        std::cerr << "unknown: " << val << "\n";
        templ->list[n].nID = 1;
      }
    }
  }
}

void Export::run(std::vector<TreeItem*> list)
{
  unsigned int       size = list.size();
  SimpleXMLTransfer* meta;
  SimpleXMLTransfer* ldata;
  int                type = 0;
  int                nPrevGroup = -1;

  // --- Read html templates ----------------------------------------
  {
    // search most local template (usr, home)
    std::string path = BaseTools::getFilePath("share/fotoalbum/templates/default/group.html");
    std::string templpath;
    if (path.length() == 0)
    {
      std::cout << "html template not found\n";
      std::cerr << "html template not found\n";
      templpath = "/home/jens/projekte/fotoalbum/CPP001/repo_fotoalbum/templates/default/";
    }
    else
      templpath = path.substr(0, path.length()-10);

    std::cout << "Using " << templpath << "\n";

    {
      std::ifstream in;
      in.open((templpath+"head.html").c_str());
      if (!in)
      {
        error = "Error opening head.html";
        return;
      }
      char inchar;
      while (!in.eof())
      {
        inchar = in.get();
        if (!in.eof())
          templ_head += inchar;
      }
      in.close();
    }

    try
    {
      templ_pic   = new Template(templpath + "picture.html");
      templ_group = new Template(templpath + "group.html");
    }
    catch (TemplateException e)
    {
      error = e.what();
      return;
    }

    PrepareTemplate(templ_group);
    PrepareTemplate(templ_pic);
  }

  // --- Prepare ----------------------------------------------------
  if (destdir[destdir.length()-1] != '/')
    destdir += '/';
  
  if (fHtml)
    pCache = new ExpPicCache(destdir);
  
  // --- Create metadata --------------------------------------------
  meta = new SimpleXMLTransfer();
  meta->setName("pictures");
  meta->setAttribute("header.version.main",             "1");
  meta->setAttribute("header.version.sub",              "1");
  meta->setAttribute("header.creator.name",             "Fotoalbum");
  meta->setAttribute("header.creator.version",          FOTOALBUMVERSION);
  meta->setAttribute("header.creator.url",              "http://kisocd.sf.net/fotoalbum.htm");
  meta->setAttribute("header.filesystem.separatorchar", "/");

  ldata = new SimpleXMLTransfer();
  ldata->setName("data");
  meta->addChild(ldata);

  error = "";
    
  std::string strThumbtable = "";
  std::string strLinks = "";
  int nFirstGroupInsideOfMe = -1;
  {
    std::string   pathOfNames = "";
    SetOfStrings  namesInDir;
    
    // separate for "linearized browsing"
    {
      std::vector<TreePicBase*>   listPics;
      std::vector<TreeGroupBase*> listGroups;
      for (unsigned int cnt=0; cnt<size; cnt++)
      {
        if (list[cnt]->getType() == TreeItem::PIC)
          listPics.push_back((TreePicBase*)(list[cnt]));
        else
          listGroups.push_back((TreeGroupBase*)(list[cnt]));
      }
      int nNr = 0;
      for (unsigned int cnt=0; cnt<listPics.size(); cnt++)
      {
        picture((TreePicBase*)(listPics[cnt]), 
                pathOfNames, namesInDir, ldata, nNr++, strThumbtable, type,
                !listGroups.empty() || cnt<listPics.size()-1,
                "index.htm");
      }
      if (!listGroups.empty())
        nFirstGroupInsideOfMe = nNr;
      for (unsigned int cnt=0; cnt<listGroups.size(); cnt++)
      {
        group((TreeGroupBase*)(listGroups[cnt]), 
              pathOfNames, namesInDir, ldata, nNr, strLinks, type,
              cnt<listGroups.size()-1,
              "index.htm", nPrevGroup);
      }            
    }
  }
  // write/close metadata file
  if (fDatafile)
  {
    std::ofstream outfile;
    std::string outfilename = destdir + "pictures.xml";
    outfile.open(outfilename.c_str());
    meta->print(outfile);
    outfile.close();
  }

  // html file
  if (fHtml)
  {
    std::string strLinkPrev, strLinkNext, strLinkUp, strURLNext;
    HtmlLinks("", -1, true, strLinkPrev, strLinkNext, strLinkUp, strURLNext);

    groupToFileHtml(destdir + "index.htm",
                    strLinkPrev, strLinkNext, strLinkUp,
                    -1, nFirstGroupInsideOfMe,
                    "Fotoalbum", "", strLinks, strThumbtable, "");
  }

  {
    bool fSchnellerHackFuerMoveExport = false;
    if (fSchnellerHackFuerMoveExport)
    {
      TemporaryTextFile* ttf;
      ttf = new TemporaryTextFile(TemporaryTextFile::GetGlobalTempDirPath() + "fotoalbummoveexport");
      if (ttf->isOK())
      {
        for (unsigned int cnt=0; cnt<copy.size(); cnt+=2)
        {
          ttf->write("mkdir -p \"" + FileTools::path(copy[cnt+1]) + "\"\n");
          ttf->write("mv \"" + copy[cnt+0] + "\" \"" + copy[cnt+1] + "\"\n\n");
        }
        ttf->close();
      }
      error = "Es ist die Option für den Move-Export aktiviert.\nDateiname:\n" + ttf->getPathAndName();
    }
  }

  delete templ_pic;
  delete templ_group;
  delete meta;
  if (fHtml)
    delete pCache;
}

int Export::run_annotate()
{
  if (!fAddToJpeg)
    return(0);
  else
  {
    TreePicBase* pic;
    int nRetCode = 0;
    std::set<std::string> kwlist;
    std::set<std::string>::iterator it;

    for (unsigned int n=0; n<annotate.size(); n++)
    {
      pic = annotate[n].ptr;
#if (EXIV2_FOUND == 1)
      try
      {
        // std::cout << "Annotate " << annotate[n].path << "\n";
        Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(annotate[n].path);
        if (image.get() != 0)
        {
          image->readMetadata();

          // exif
          {
            Exiv2::ExifData exifData = image->exifData();
            Exiv2::Exifdatum& tag = exifData["Exif.Photo.DateTimeOriginal"];
            std::string datestr = pic->getDateString();
            std::string timestr = pic->getTimeString();
            if (timestr.length() == 0)
              timestr = "00:00:00";
            if (datestr.length() == 0)
              datestr = "0000:00:00";
            else
            {
              datestr[4] = ':';
              datestr[7] = ':';
            }
            tag.setValue(datestr + " " + timestr);

            image->setExifData(exifData);
          }

# if 0
          // Doing this sometimes leads to fltk not being able to load the image, so
          // I don't export to jpeg comment. Data is in iptc caption, anyway. However,
          // every other app I tried is able to load the image (internal jpegtran is
          // able to do so, too).
          //
          // jpeg comment
          {
            if (pic->getComment().length())
              image->setComment(pic->getComment());
          }
# endif

          // iptc and xmp
          {
            TreeItem* item = pic;
            Exiv2::IptcData& iptcData = image->iptcData();
            Exiv2::XmpData&  xmpData  = image->xmpData();
            Exiv2::IptcKey ek("Iptc.Application2.Keywords");
            Exiv2::XmpKey  xk("Xmp.dc.subject");

            // read existing keywords into kwlist
            kwlist.clear();
            if (fKwMerge)
              ExifInfo::ReadKeywords(image, kwlist);

            // delete all existing keywords
            if (!iptcData.empty())
            {
              Exiv2::IptcData::iterator md  = iptcData.begin();
              while (md != iptcData.end())
              {
                if (md->key().compare("Iptc.Application2.Keywords") == 0)
                  md = iptcData.erase(md);
                else
                  md++;
              }
            }

            // delete all existing keywords
            if (!xmpData.empty())
            {
              Exiv2::XmpData::iterator md  = xmpData.begin();
              while (md != xmpData.end())
              {
                if (md->key().compare("Xmp.dc.subject") == 0)
                  md = xmpData.erase(md);
                else
                  md++;
              }
            }

            // get keywords
            item->kw_set(kwlist);

            if (fKwInherit)
            {
              // inherit keywords of groups above
              item = item->itemParent;
              while (item)
              {
                // By using kw_list() the order of keywords is preserved, but duplicates are
                // possible. Using kw_set() removes duplicates but changes order, but it gives
                // an easy way to merge with keywords which are in the file already.
                item->kw_set(kwlist);
                item = item->itemParent;
              }
            }

            //
            it = kwlist.begin();
            Exiv2::Value::AutoPtr vx = Exiv2::Value::create(Exiv2::xmpBag);
            while (it != kwlist.end())
            {
              Exiv2::Value::AutoPtr v = Exiv2::Value::create(Exiv2::string);
              v-> read(*it);
              vx->read(*it);
              iptcData.add(ek, v.get ());
              it++;
            }
            xmpData.add(xk, vx.get());

            // caption
            if (pic->getComment().length())
            {
              // set comment if it is allowed to be overwritten or if there is no comment yet
              Exiv2::Iptcdatum& caption = iptcData["Iptc.Application2.Caption"];
              std::string captionstr = trim(caption.toString());

              if (fCommentOverwrite || captionstr.length() == 0)
                iptcData["Iptc.Application2.Caption"] = pic->getComment();
            }

            //
            image->setIptcData(iptcData);
          }

          //
          image->writeMetadata();
        }
      }
      catch (Exiv2::AnyError& e)
      {
        std::cerr << "Caught Exiv2 exception '" << e << "'\n";
        nRetCode = -1;
      }
#endif
    }

    return(nRetCode);
  }
}

ExpPicCache::ExpPicCache(std::string strDir)
{
  this->strDir = strDir;
  try
  {
    xState = new SimpleXMLTransfer(strDir + "/content.xml");
  }
  catch (XMLException e)
  {
    std::cerr << "Error loading export cache state: " << e.what() << "\n";
    xState = new SimpleXMLTransfer();
  }
  
  nNum = 0;
  
  for (int n=0; n<xState->getChildCount(); n++)
  {
    SimpleXMLTransfer* xItem = xState->getChildAt(n);
    mapC.insert(std::make_pair(xItem->attribute("n"), xItem));
    
    int nNr = xItem->attributeAsInt("i");
    if (nNum <= nNr)
      nNum = nNr+1;    
  }
}

ExpPicCache::~ExpPicCache()
{
  std::ofstream out;
  out.open((strDir + "/content.xml").c_str());
  xState->print(out, 0, 2);
  out.close();
  delete xState;
}

std::string ExpPicCache::UpToDate(std::string strSrc, std::string strDstDir,
                                  int nW, int nH, int nMode, int nQual)
{
  std::map<std::string, SimpleXMLTransfer*>::iterator it = mapC.find(strSrc);
  if (it == mapC.end())
    return("");
  
  std::string strName = Name(strSrc, nW, nH, nMode, nQual, it->second->attributeAsInt("i"));
  if (FileTools::isFileUpToDate(strSrc, strDstDir +"/"+strName) > 0)
    return(strName);
  else
    return("");
}

std::string ExpPicCache::Insert(std::string strSrc, std::string strDstDir,
                                int nW, int nH, int nMode, int nQual)
{
  SimpleXMLTransfer* xP;
  int nRes;
  std::map<std::string, SimpleXMLTransfer*>::iterator it = mapC.find(strSrc);
  if (it == mapC.end())
  {
    xP = new SimpleXMLTransfer();
    xP->setName("p");
    xP->addAttribute("n", strSrc);
    nRes = nNum++;
    xP->addAttribute("i", nRes);
    xState->addChild(xP);
    mapC.insert(std::make_pair(strSrc, xP));
  }
  else
  {
    xP = it->second;
    nRes = xP->attributeAsInt("i");
  }

  std::string strName = Name(strSrc, nW, nH, nMode, nQual, nRes);

  return(strName);
}
 
std::string ExpPicCache::Name(std::string strSrc, int nW, int nH, int nMode, int nQual,
                              int nNr)
{
  return("p/" + itoHexStr(nNr, 16) + "_" 
         + itoStr(nW, ' ', 1) + itoStr(nH, ' ', 1)
         + "_" + itoStr(nQual, '0', 3) + itoHexStr(nMode, 4) + FileTools::getSuffix(strSrc));
}

