/*
 * Copyright (C) 2005  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#ifndef LINKS_H
#define LINKS_H

#include <vector>

#include "base_defs.h"
#include "treeitem.h"

class   TreeView;
class   LinkableItemInfo;
typedef std::vector<LinkableItemInfo> LinkInfoList;


/**
 * Various errors, may be or'ed together:
 */

// link ID could not be found or has wrong type. Not critical.
#define eLiErrLink         0x08

// Link count underflow. CRITICAL!
#define eLiErrUnderflow    0x10

// Link ID could not be found. CRITICAL!
#define eLiErrIDNotInList  0x20

class LinkInfo
{
  public:
   
   /**
    * 
    */
   LinkInfo(TreeView* myparent);
   
   /**
    * Inserts a new item into the sorted list
    */
   void newLinkableItem(TreeItem* item);

   /**
    * 
    */
   int removeFromList(unsigned int uID);
   
   /**
    * Returns a pointer to the element with uID
    */
   TreeItem* getItemWithID(unsigned int uID);
   
   /**
    * 
    */
   void showList();
   
   /**
    * Inc/Decrements link count of element with uID == uIDToDecr.
    * Returns 0 if everything is fine, 2 on link count underflow,
    * 1 if the link source is not there (or has a wrong type).
    */
   int countLinkCount(TreeItem**            ptr,
                      unsigned int          uIDToCount,
                      TreeItem::eItemType   expectedType,                      
                      bool                  fIncr);
      
   /**
    * 
    */
   unsigned int getLinkCnt(unsigned int uID);
   
   /**
    * 
    */
   void clear();
   
   /**
    * 
    */
   unsigned int getIDMax();
   
  private:
   
   /**
    * Returns the index of the item with uID or of the first one with
    * a bigger uID.
    */
   unsigned int getIndexNear(unsigned int uID);

   /**
    * 
    */
   LinkInfoList lil;
   
   /**
    * 
    */
   TreeView* parent;
};

/**
 * Contains information about linkable items or provides access to them -- 
 * depends on compile time settings.
 * 
 * uLinkCnt: how many links point to the item? Links from comments are only
 * counted if the comment is real data. This is important when data is loaded
 * initially and when a link is removed.
 * 
 * When a comment is edited, one does not have to care about whether this 
 * comment is real data or linked -- in any case, real data is changed, so 
 * link count should be decreased before editing and increased afterwards 
 * regardless of the item's type.
 * 
 * See @ref TreeView::getLinkInfoList and use it, if needed.
 */
class LinkableItemInfo
{
  public:
      
   /**
    * Returns the ID of the linkable item
    */
   inline unsigned int getID() 
   { 
#if BASE_ID_IN_LIST == 1
     return(uID);
#else 
     return(ptr->uID);
#endif   
   }
   
   /**
    * Sets the ID of the linkable item
    */
   inline void setID(unsigned int uNewID)
   { 
#if BASE_ID_IN_LIST == 1
     uID = uNewID;
#endif
   }
      
   /**
    * Pointer to the linkable item
    */
   TreeItem*    ptr;
            
   /**
    * Returns the link count of the linkable item
    */
   inline unsigned int getLinkCnt()
   { 
#if BASE_LNKCNT_IN_LIST == 1
     return(uLinkCnt);
#else 
     return(ptr->uLinkCnt);
#endif   
   }
   
   /**
    * Sets the link count of the linkable item
    */
   inline void setLinkCnt(unsigned int uNewLC)
   { 
#if BASE_LNKCNT_IN_LIST == 1
     uLinkCnt = uNewLC;
#else     
     ptr->uLinkCnt = uNewLC;
#endif   
   }   
   
  private:
   
#if BASE_ID_IN_LIST == 1   
   /**
    * ID of linkable item
    */
   unsigned int uID;
#endif   

#if BASE_LNKCNT_IN_LIST == 1
   /**
    * How many links point to this item?
    */
   unsigned int uLinkCnt;
#endif   
};

#endif 
