// -*- mode: c; mode: fold -*-
//
/*
 * Copyright (C) 2005-2008, 2010  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "treegroupbase.h"

#include <iostream>

#include <sys/stat.h>
#include <unistd.h>

#include "../shared/scandir.h"
#include "../shared/filetools.h"
#include "../shared/lib_conversions.h"
#include "xmldefs.h"
#include "treegroupdata.h"
#include "treegrouplinkproperties.h"
#include "treepicdata.h"
#include "treepiclink.h"

TreeGroupBase::TreeGroupBase(TreeItem* prev, TreeGroupBase* parent) : TreeItem(prev, parent)
{
  if (parent == (TreeItem*)0)
    fExpanded = true;
  else
    fExpanded = false;

  itemFirstChild = (TreeItem*)0;
  nGroups        = 0;
}

TreeGroupBase::~TreeGroupBase()
{
  deleteChildren();
}

std::string TreeGroupBase::getSomeText(eText whichText)                             
{
  return("");
}


int  TreeGroupBase::loadFromXML_children(SimpleXMLTransfer*    source,
                                         LinkInfo*             links)           
{
  int nSize = source->getChildCount();

  TreeItem*          itemdst;
  SimpleXMLTransfer* itemsrc;
  std::string        lType;
  TreeItem*          prev = (TreeItem*)0;

  itemFirstChild = (TreeItem*)0;
  nGroups        = 0;

  for (int n=0; n<nSize; n++)
  {
    itemdst = (TreeItem*)0;
    itemsrc = source->getChildAt(n);
    lType   = itemsrc->attribute(XMLNAME_TYPE);

    if (lType.compare(XMLNAME_TYPE_GROUP) == 0)
    {
      itemdst = new TreeGroupData(prev, this);
      nGroups++;
    }
    else if (lType.compare(XMLNAME_TYPE_GROUPLINKPROPERTIES) == 0)
    {
      itemdst = new TreeGroupLinkProperties(prev, this);
      nGroups++;
    }
    else if (lType.compare(XMLNAME_TYPE_PIC) == 0)
    {
      itemdst = new TreePicData(prev, this);
    }
    else if (lType.compare(XMLNAME_TYPE_PICLINK) == 0)
    {
      itemdst = new TreePicLink(prev, this);
    }
    
    if (itemdst != (TreeItem*)0)
    {
      itemdst->loadFromXML_all(itemsrc, links);
      
      if (lType.compare(XMLNAME_TYPE_PIC) == 0)
        ((TreePicData*)itemdst)->setHidden();
      
      prev = itemdst;
    }        
  }

  return(0);
}


int  TreeGroupBase::saveToXML_children(SimpleXMLTransfer* dest, bool fConvLinks, int nVersion)
{
  TreeItem*          itemsrc = itemFirstChild;
  SimpleXMLTransfer* itemdst;

  while (itemsrc != (TreeItem*)0)
  {
    itemdst = new SimpleXMLTransfer();

    itemsrc->saveToXML_all(itemdst, fConvLinks, nVersion);
    dest->addChild(itemdst);

    itemsrc = itemsrc->itemNext;
  }

  return(0);
}


int  TreeGroupBase::deleteChildren()                                            
{
  TreeItem* p = itemFirstChild;
  TreeItem* q;

  while (p != (TreeItem*)0)
  {
    q = p;
    p = p->itemNext;
    delete q;
  }
  
  itemFirstChild = 0;
  
  return(0);
}


enum TreeItem::eItemType TreeGroupBase::getType()                               
{
  return(GROUP);
}


TreePicBase* TreeGroupBase::firstPicture()                                      
{
  TreeItem* p = itemFirstChild;
  TreePicBase* item;

  while (p != (TreeItem*)0)
  {
    switch (p->getType())
    {
     case PIC:
      return((TreePicBase*)p);
      break;
     case GROUP:
      item = ((TreeGroupBase*)p)->firstPicture();
      if (item != (TreePicBase*)0)
        return(item);
      break;
     default:
      break;
    }
    p = p->itemNext;
  }
  return((TreePicBase*)0);
}


void TreeGroupBase::unmarkAll()                                                 
{
  std::cout << "void TreeGroupBase(" << this << ")::unmarkAll()\n";

  TreeItem* p = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    p->fMarked = false;
    p = p->itemNext;
  }
}


bool TreeGroupBase::canExpand(bool fShowPics)                                   
{
  if (nGroups)
    return(true);
  else
  {
    // There are no groups inside of me.    
    if (fShowPics && itemFirstChild != (TreeItem*)0)
    {
      // So there may be a picture inside of me. There must be one 
      // which is not hidden!
      TreeItem* p = itemFirstChild;
      
      while (p != (TreeItem*)0 && p->isHidden())
        p = p->itemNext;
      
      if (p != (TreeItem*)0)
        return(true);
      else
        return(false);
    }
    else
      return(false);
  }
}


int  TreeGroupBase::connectLinks(LinkInfo* links)                               
{
  int       nRetCode = 0;
  TreeItem* p        = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    nRetCode |= p->connectLinks(links);

    p = p->itemNext;
  }
  return(nRetCode);
}


TreeItem* TreeGroupBase::getLastChild()                                         
{
  TreeItem* p    = itemFirstChild;
  TreeItem* pold = p;

  while (p != (TreeItem*)0)
  {
    pold = p;
    p    = p->itemNext;
  }

  return(pold);
}


TreeGroupBase* TreeGroupBase::addNewGroup(const char*   name,
                                          eItemLinkType linktype)               
{
  TreeGroupBase* ng = (TreeGroupBase*)0;

  nGroups++;

  switch (linktype)
  {
   case DATA:
    ng = new TreeGroupData(getLastChild(), this);
    ng->setTitle(name);
    break;
   case LINKPROPERTIES:
    ng = new TreeGroupLinkProperties(getLastChild(), this);
    break;
   default:
    break;
  }

  return(ng);
}


TreePicBase*   TreeGroupBase::addNewPic(eItemLinkType linktype)                 
{
  TreePicBase* np = (TreePicBase*)0;

  switch (linktype)
  {
   case DATA:
    np = new TreePicData(getLastChild(), this);
    break;
   case LINK:
    np = new TreePicLink(getLastChild(), this);
    break;
   default:
    break;
  }

  return(np);
}


void TreeGroupBase::changeBasepath(std::string oldbp, std::string newbp)        
{
  TreeItem* p    = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    p->changeBasepath(oldbp, newbp);
    p = p->itemNext;
  }
}


void TreeGroupBase::addPics(std::string                    srcdir,
                            std::string                    bp,
                            TreePicData::eDateTimeSource   eDTS)                
{
  ScanDir*        dir;
  struct stat     Attribs;              // ein Puffer für die Dateiattribute

  std::cout << getText() << ": adding pictures from " << srcdir << "\n";

  dir = new ScanDir(srcdir.c_str());

  for (int n=0; n<dir->nNrEntries; n++)
  {
    // read Attribs
    {
      std::string tmp = srcdir;

      tmp += (dir->list)[n]->d_name;
      if (tmp[tmp.length()-1] == '/')
        tmp = tmp.substr(0, tmp.length()-1);
	  
      stat(tmp.c_str(), &Attribs);
    }

    if (S_ISDIR(Attribs.st_mode) != 0)
    {
      // Es ist ein Verzeichnis!
      std::string tmp = (dir->list)[n]->d_name;
      if (tmp[tmp.length()-1] == '/')
        tmp = tmp.substr(0, tmp.length()-1);

      if ((tmp.compare("..") != 0) &&
          (tmp.compare(".")  != 0) &&
          (tmp.compare(".thumbnails") != 0) && // thumbnail directory used by fotoalbum
          (tmp.compare(".thumbs")     != 0)    // thumbnail directory used by mapivi
          )
      {
        std::cout << "directory: " << tmp << "\n";

        TreeGroupBase* ng = addNewGroup(tmp.c_str(), DATA);
        ng->addPics(srcdir + tmp + "/", bp, eDTS);
      }
    }
    else if (S_ISREG(Attribs.st_mode) != 0)
    {
      std::string tmp = (dir->list)[n]->d_name;
      
      if ((strU(tmp).compare("THUMBS.DB") != 0))
      {
        TreePicData* pic = (TreePicData*)addNewPic(DATA);
        pic->setData(srcdir + (dir->list)[n]->d_name, bp, eDTS);
        pic->setHidden();
      }
    }
  }
  delete dir;
}


void TreeGroupBase::addFile(std::string                    srcdir,
                            std::string                    bp,
                            TreePicData::eDateTimeSource   eDTS)                
{
  TreePicData* pic = (TreePicData*)addNewPic(DATA);
  pic->setData(srcdir, bp, eDTS);
  pic->setHidden();
}


TreeItem* TreeGroupBase::linkAfter(TreeItem*    item,
                                   TreeItem*    dest,
                                   unsigned int uLinkID)                        
{
  // where to put it?
  TreeItem* prev = itemFirstChild;
  TreeItem* next = (TreeItem*)0;

  if (dest == (TreeItem*)0)
  {
    // first child
    prev = (TreeItem*)0;
    next = itemFirstChild;
  }
  else if (prev != (TreeItem*)0)
  {
    while (prev != (TreeItem*)0 && prev != dest)
      prev = prev->itemNext;

    if (prev != dest)
    {
      // first child
      prev = (TreeItem*)0;
      next = itemFirstChild;
    }
    else
    {
      next = prev->itemNext;
    }
  }

  // real item has to have an ID:
  item->setID(uLinkID);

  // create item
  TreeItem* newItem;

  switch (item->getType())
  {
   case TreeItem::GROUP:
    newItem = new TreeGroupLinkProperties(prev, this, item);
    nGroups++;
    break;
   case TreeItem::PIC:
    newItem = new TreePicLink(prev, this, item);
    break;
   default:
    std::cerr << "Internal error\n";
    return((TreeItem*)0);
    break;
  }

  // insert into tree
  if (next != (TreeItem*)0)
  {
    newItem->itemNext  = next;
    next->itemPrev     = newItem;
  }

  return(newItem);
}


void TreeGroupBase::moveAfter(TreeItem* item, TreeItem* dest)                   
{
  TreeGroupBase* parent = (TreeGroupBase*)item->itemParent;
  if (parent != (TreeGroupBase*)0)
    parent->removeItem(item);

  item->itemParent = this;

  if (item->getType() == TreeItem::GROUP)
    nGroups++;

  TreeItem* prev = itemFirstChild;

  if (prev == (TreeItem*)0)
  {
    // There is only one place to put the item.
    item->itemPrev = (TreeItem*)0;
    item->itemNext = (TreeItem*)0;
    itemFirstChild = item;
  }
  else
  {
    while (prev != (TreeItem*)0 && prev != dest)
      prev = prev->itemNext;

    if (prev != dest || dest == (TreeItem*)0)
    {
      // first child
      item->itemNext = itemFirstChild;
      itemFirstChild->itemPrev = item;
      item->itemPrev = (TreeItem*)0;
      itemFirstChild = item;
    }
    else
    {
      TreeItem* next = prev->itemNext;

      item->itemNext = next;
      if (next != (TreeItem*)0)
        next->itemPrev = item;

      prev->itemNext = item;
      item->itemPrev = prev;
    }
  }
}


void TreeGroupBase::removeItem(TreeItem* item)                                  
{
  TreeItem* p = itemFirstChild;

  while (p != (TreeItem*)0 && p != item)
    p = p->itemNext;

  if (p == item)
  {
    bool      fPics;
    TreeItem* prev = item->itemPrev;
    TreeItem* next = item->itemNext;

    item->itemParent = (TreeGroupBase*)0;

    if (prev != (TreeItem*)0)
      prev->itemNext = next;
    if (next != (TreeItem*)0)
      next->itemPrev = prev;
    if (itemFirstChild == item)
      itemFirstChild = next;

    // Had pictures been shown? I can only guess...
    // In case pictures are shown and there are groups inside of
    // me, I will collapse, although I could stay expanded.
    if (nGroups == 0 && fExpanded)
      fPics = true;
    else
      fPics = false;

    if (item->getType() == TreeItem::GROUP)
      nGroups--;

    if (fExpanded && canExpand(fPics) == false)
      fExpanded = false;
  }
}


unsigned int TreeGroupBase::getMinMaxDateTime(JWWDateTime& min,
                                              JWWDateTime& max,
                                              bool         fRecursive)          
{
  TreeItem*     ptr = itemFirstChild;
  unsigned int  nNonGroup = 0;

  while (ptr != (TreeItem*)0)
  {
    if (ptr->getType() == TreeItem::GROUP && fRecursive)
    {
      ((TreeGroupBase*)ptr)->getMinMaxDateTime(min, max, fRecursive);
    }
    else if (ptr->getType() == TreeItem::PIC && !ptr->isHidden())
    {
      nNonGroup++;

      if (((TreePicBase*)ptr)->hasDate())
      {
        JWWDateTime dt = ptr->getDT();

        if (!min.isDateSet())
          min = dt;
        else if (min.compare(&dt) < 0)
          min = dt;

        if (!max.isDateSet())
          max = dt;
        else if (max.compare(&dt) > 0)
          max = dt;
      }
    }

    ptr = ptr->itemNext;
  }

  return(nNonGroup);
}


void TreeGroupBase::sortChildrenChrono()                                       
{
  bool      fChanged;
  bool      fEnd;
  TreeItem* p1    = itemFirstChild;
  TreeItem* p2;
  int       nC;

  std::cout << "void TreeGroupBase::sortChildrenChrono()\n";
  std::cout.flush();

  do
  {
    fChanged = false;
    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemNext;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else
      {
        JWWDateTime dt = p2->getDT();
        nC = p1->getDT().compare( &dt );

        if (nC == 1)
        {
          fChanged = true;

          TreeItem* p0 = p1->itemPrev;
          TreeItem* p3 = p2->itemNext;
          // p0 -> p1 -> p2 -> p3
          //
          if (p0 != (TreeItem*)0)
            p0->itemNext = p2;
          p2->itemPrev = p0;
          p2->itemNext = p1;
          p1->itemPrev = p2;
          p1->itemNext = p3;
          if (p3 != (TreeItem*)0)
            p3->itemPrev = p1;
        }
        p1 = p2;
      }
    }

    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemPrev;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else
      {
        JWWDateTime dt = p2->getDT();
        nC = p1->getDT().compare( &dt );

        if (nC == -1)
        {
          fChanged = true;
          TreeItem* p3 = p2->itemPrev;
          TreeItem* p0 = p1->itemNext;
          // p3 -> p2 -> p1 -> p0
          //
          if (p3 != (TreeItem*)0)
            p3->itemNext = p1;
          p1->itemPrev = p3;
          p1->itemNext = p2;
          p2->itemPrev = p1;
          p2->itemNext = p0;
          if (p0 != (TreeItem*)0)
            p0->itemPrev = p2;
        }
        p1 = p2;
      }
    }
  }
  while (fChanged);

  itemFirstChild = p1;
}

void TreeGroupBase::sortChildrenPath()
{
  bool      fChanged;
  bool      fEnd;
  TreeItem* p1    = itemFirstChild;
  TreeItem* p2;
  int       nC;

  do
  {
    fChanged = false;
    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemNext;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else
      {
        nC = p1->getText().compare(p2->getText());

        if (nC == 1)
        {
          fChanged = true;

          TreeItem* p0 = p1->itemPrev;
          TreeItem* p3 = p2->itemNext;
          // p0 -> p1 -> p2 -> p3
          //
          if (p0 != (TreeItem*)0)
            p0->itemNext = p2;
          p2->itemPrev = p0;
          p2->itemNext = p1;
          p1->itemPrev = p2;
          p1->itemNext = p3;
          if (p3 != (TreeItem*)0)
            p3->itemPrev = p1;
        }
        p1 = p2;
      }
    }

    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemPrev;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else
      {
        nC = p1->getText().compare(p2->getText());

        if (nC == -1)
        {
          fChanged = true;
          TreeItem* p3 = p2->itemPrev;
          TreeItem* p0 = p1->itemNext;
          // p3 -> p2 -> p1 -> p0
          //
          if (p3 != (TreeItem*)0)
            p3->itemNext = p1;
          p1->itemPrev = p3;
          p1->itemNext = p2;
          p2->itemPrev = p1;
          p2->itemNext = p0;
          if (p0 != (TreeItem*)0)
            p0->itemPrev = p2;
        }
        p1 = p2;
      }
    }
  }
  while (fChanged);

  itemFirstChild = p1;
}

void TreeGroupBase::sortChildrenSize(std::string bp)
{
  bool      fChanged;
  bool      fEnd;
  TreeItem* p1    = itemFirstChild;
  TreeItem* p2;

  do
  {
    fChanged = false;
    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemNext;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else if (dynamic_cast<TreePicBase*>(p1) != 0 && dynamic_cast<TreePicBase*>(p2) != 0)
      {
        if (dynamic_cast<TreePicBase*>(p1)->GetFileSize(bp) > dynamic_cast<TreePicBase*>(p2)->GetFileSize(bp))
        {
          fChanged = true;

          TreeItem* p0 = p1->itemPrev;
          TreeItem* p3 = p2->itemNext;
          // p0 -> p1 -> p2 -> p3
          //
          if (p0 != (TreeItem*)0)
            p0->itemNext = p2;
          p2->itemPrev = p0;
          p2->itemNext = p1;
          p1->itemPrev = p2;
          p1->itemNext = p3;
          if (p3 != (TreeItem*)0)
            p3->itemPrev = p1;
        }
        p1 = p2;
      }
    }

    fEnd     = false;
    while (!fEnd)
    {
      p2 = p1->itemPrev;

      if (p2 == (TreeItem*)0)
      {
        fEnd = true;
      }
      else if (dynamic_cast<TreePicBase*>(p1) != 0 && dynamic_cast<TreePicBase*>(p2) != 0)
      {
        if (dynamic_cast<TreePicBase*>(p1)->GetFileSize(bp) < dynamic_cast<TreePicBase*>(p2)->GetFileSize(bp))
        {
          fChanged = true;
          TreeItem* p3 = p2->itemPrev;
          TreeItem* p0 = p1->itemNext;
          // p3 -> p2 -> p1 -> p0
          //
          if (p3 != (TreeItem*)0)
            p3->itemNext = p1;
          p1->itemPrev = p3;
          p1->itemNext = p2;
          p2->itemPrev = p1;
          p2->itemNext = p0;
          if (p0 != (TreeItem*)0)
            p0->itemPrev = p2;
        }
        p1 = p2;
      }
    }
  }
  while (fChanged);

  itemFirstChild = p1;
}

JWWDateTime TreeGroupBase::getDT()                                              
{
  JWWDateTime min;
  JWWDateTime max;

  min.nTime = 0;
  min.nDate = 0;
  max = min;

  getMinMaxDateTime(min, max, true);

  return(min);
}


unsigned int TreeGroupBase::countMissingFiles(std::string basepath)             
{
  unsigned int uCnt = 0;
  TreeItem*    p    = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    uCnt += p->countMissingFiles(basepath);
    p     = p->itemNext;
  }

  return(uCnt);
}


std::string TreeGroupBase::findFile(std::string basepath,
                                    std::string filename)                       
{
  TreeItem* p = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    if (p->getType() == TreeItem::GROUP)
    {
      std::string erg = ((TreeGroupBase*)p)->findFile(basepath, filename);
      if (erg.length())
        return(erg);
    }
    else
    {
      std::string path = basepath + p->getText();

      path = path.substr(0, path.rfind('/')+1) + filename;

      if (FileTools::fileExists(path))
        return(path);
    }

    p = p->itemNext;
  }

  return("");
}


void TreeGroupBase::linkPicsInto(TreeGroupBase* dest)                           
{
  TreeItem* p       = itemFirstChild;
  TreeItem* newItem = (TreeItem*)0;

  while (p != (TreeItem*)0)
  {
    if (p->getType() == TreeItem::PIC)
    {
      if (p->isHidden() == false)
        newItem = dest->linkAfter(p, newItem, 0);
    }
    else
    {
      ((TreeGroupBase*)p)->linkPicsInto(dest);
    }

    p = p->itemNext;
  }
}


void TreeGroupBase::deleteDuplicates()                                          
{
  TreePicLink* p = (TreePicLink*)itemFirstChild;
  TreePicLink* q;

  std::cout << "void TreeGroupBase::deleteDuplicates()\n";
  std::cout.flush();

  while (p != (TreePicLink*)0)
  {
    q = (TreePicLink*)p->itemNext;

    if (q != (TreePicLink*)0)
    {
      if (p->getDataSource() == q->getDataSource())
      {
        removeItem(q);
        delete q;
      }
    }

    p = (TreePicLink*)p->itemNext;
  }
}


void TreeGroupBase::subdivideChrono(int level)                                  
{
  TreePicBase*   p      = (TreePicBase*)itemFirstChild;
  TreeGroupBase* group  = (TreeGroupBase*)0;
  TreePicBase*   ptemp;
  TreeItem*      lastItemInGroup = (TreeItem*)0;

  std::cout << "void TreeGroupBase::subdivideChrono(" << level << ")  " << getText() << "\n";
  std::cout.flush();

  // --- subdivide by year ------------------------------------------
  if (level == 0)
  {
    unsigned int   year   = 9999;

    while (p != (TreePicBase*)0)
    {
      ptemp = (TreePicBase*)p->itemNext;

      if (p->getYear() != year)
      {
        std::string newname;

        year = p->getYear();

        if (year != 0)
          newname = itoStr(year, ' ', 1);
        else
          newname = "no date";

        //
        if (group != (TreeGroupBase*)0)
          group->subdivideChrono(1);

        // create a new group
        TreeGroupData* newGroup  = new TreeGroupData(group, this);
        nGroups++;
        newGroup->setTitle(newname.c_str());
        p->itemPrev     = newGroup;
        group           = newGroup;
        group->itemNext = p;
        lastItemInGroup = (TreeItem*)0;
      }

      group->moveAfter(p, lastItemInGroup);
      lastItemInGroup = p;

      p = ptemp;
    }

    if (group != (TreeGroupBase*)0)
      group->subdivideChrono(1);
  }
  // --- subdivide --------------------------------------------------
  else
  {
    unsigned int uCnt = 0;

    while (p != (TreePicBase*)0)
    {
      ptemp = (TreePicBase*)p->itemNext;

      if (uCnt == 0)
      {
        uCnt = 200-1;

        std::string newname;

        if (p->hasDate())
          newname = p->getDateString() + " " + p->getTimeString();
        else
          newname = "no date";

        // create a new group
        TreeGroupData* newGroup  = new TreeGroupData(group, this);
        nGroups++;
        newGroup->setTitle(newname.c_str());
        p->itemPrev     = newGroup;
        group           = newGroup;
        group->itemNext = p;
        lastItemInGroup = (TreeItem*)0;
      }
      else
        uCnt--;

      group->moveAfter(p, lastItemInGroup);
      lastItemInGroup = p;

      p = ptemp;
    }
  }

}


void TreeGroupBase::setMarkRec(bool fVal)                                       
{
  fMarked = fVal;

  TreeItem*    p    = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    if (p->getType() == TreeItem::PIC)
      p->fMarked = fVal;
    else
      ((TreeGroupBase*)p)->setMarkRec(fVal);

    p = p->itemNext;
  }
}


bool TreeGroupBase::markMatchingItems(SearchData sd, TreeView* DataToMark)
{
  bool      fPicMatched = false;
  
  if ((sd.nFlags & SD_GROUP) != 0)
  {
    if ((sd.nFlags & SD_KEYWORDS) != 0)
    {
      sd.removeMatchingWords(getKeywords());
    }
    if ((sd.nFlags & SD_COMMENT) != 0)
    {
      sd.removeMatchingWords(getComment());
    }
    if ((sd.nFlags & SD_TITLE) != 0)
    {
      sd.removeMatchingWords(getText());
    }
  }

  if (sd.match() && DataToMark == 0)
  {
    fMarked = true;
    setMarkRec(true);
  }
  else
  {
    fMarked = false;

    TreeItem*    p = itemFirstChild;

    while (p != (TreeItem*)0)
    {
      if (p->getType() == TreeItem::PIC)
      {
        if (((TreePicBase*)p)->markIfMatch(sd, DataToMark))
        {
          fPicMatched = true;
        }
      }
      else if (DataToMark == 0)
      {
        if (((TreeGroupBase*)p)->markMatchingItems(sd))
          fMarked = true;
      }

      p = p->itemNext;
    }
  }
  
  if (fPicMatched && DataToMark == 0)
    fMarked = true;
  
  if (DataToMark == 0)
    return(fMarked);
  else
    return(fPicMatched);
}


bool TreeGroupBase::markItemsWithoutText()                                      
{
  fMarked = false;

  TreeItem*    p = itemFirstChild;

  while (p != (TreeItem*)0)
  {
    if (p->getType() == TreeItem::PIC)
    {
      if (((TreePicBase*)p)->markIfWithoutText())
        fMarked = true;
    }
    else
    {
      if (((TreeGroupBase*)p)->markItemsWithoutText())
        fMarked = true;
    }

    p = p->itemNext;
  }

  return(fMarked);
}


void TreeGroupBase::linkMarkedItems(TreeGroupBase* src)                         
{
  TreeItem* p       = src->itemFirstChild;
  TreeItem* newItem = (TreeItem*)0;

  while (p != (TreeItem*)0)
  {
    if (p->fMarked)
    {
      newItem = linkAfter(p, newItem, 0);

      if (p->getType() == GROUP)
      {
        ((TreeGroupBase*)newItem)->linkMarkedItems((TreeGroupBase*)p);
      }
    }

    p = p->itemNext;
  }
}


void TreeGroupBase::collapsePicsHidden()                                        
{
  if (fExpanded)
  {
    TreeItem* p = itemFirstChild;

    while (p != (TreeItem*)0)
    {
      if (p->getType() == GROUP)
        ((TreeGroupBase*)p)->collapsePicsHidden();

      p = p->itemNext;
    }

    if (canExpand(false) == false)
      fExpanded = false;
  }
}


void TreeGroupBase::kw_setGroup(std::set<std::string>& kwset)
{
  // myself:
  kw_set(kwset);
  
  // children:
  TreeItem*    p = itemFirstChild;
  while (p != (TreeItem*)0)
  {
    if (p->getType() == GROUP)
      ((TreeGroupBase*)p)->kw_setGroup(kwset);
    else    
      p->kw_set(kwset);
    
    p = p->itemNext;
  }
}

void TreeGroupBase::kw_replaceGroup(std::string from, std::string to, int mode)
{
  // myself:
  kw_replace(from, to, mode);
  
  // children:
  TreeItem* p = itemFirstChild;  
  while (p != (TreeItem*)0)
  {
    if (p->getType() == GROUP)
      ((TreeGroupBase*)p)->kw_replaceGroup(from, to, mode);
    else
      p->kw_replace(from, to, mode);
    
    p = p->itemNext;
  }  
}

void TreeGroupBase::countNonGroupItems(unsigned int& inside, unsigned int& insideandbelow)
{
  TreeItem*    p = itemFirstChild;
  unsigned int dummy;
  
  while (p != (TreeItem*)0)
  {
    if (p->getType() == GROUP)    
      ((TreeGroupBase*)p)->countNonGroupItems(dummy, insideandbelow);
    else
    {
      inside++;
      insideandbelow++;
    }
    
    p = p->itemNext;
  }  
}

void TreeGroupBase::RecursiveList(std::set<TreeItem*>& set)
{
  TreeItem* p = itemFirstChild;
  
  while (p != (TreeItem*)0)
  {
    p->RecursiveList(set);
    p = p->itemNext;
  }
  
  set.insert(this);    
}
