/*
 * Copyright (C) 2005, 2008-2010  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#ifndef TREEITEM_H
# define TREEITEM_H

# include <set>
# include "../shared/SimpleXMLTransfer.h"
# include "base_defs.h"
# include "datetime.h"

class LinkInfo;
class TreeGroupBase;

class TreeItem
{
  friend class LinkableItemInfo;
  
public:
  enum eItemType { GROUP, PIC, ANY, NONE };
  enum eItemLinkType { ROOTDATA, DATA, LINKPROPERTIES, LINK };
  enum eText { TEXT_EXIF };
  
public:
  
  /**
   * 
   */
  TreeItem(TreeItem* prev, TreeGroupBase* parent);
  
  
  virtual ~TreeItem();
  
  /**
   * This is the mark which can be seen/toggled by the user in the GUI.
   * 
   * It is public to allow for fast reset in a whole group.
   * Die Markierung wird benutzt bei:
   * -löschen
   * -drehen von Bildern
   * -Zeitverschiebung
   * -Verschieben/Verlinken von Elementen
   * -Elementen einen bestimmten Text hinzufügen (bei keywords oder
   *  comment).
   * - ...
   */
  bool fMarked;
  
  /**
   *
   */
  virtual std::string getText() = 0;
  
  /**
   * 
   */
  virtual std::string getKeywords() = 0;
  
  /**
   * 
   */
  virtual std::string getComment() = 0;
  
  /**
   * Returns comment as html. No matter which format it has been written in,
   * it will be returned as html -- at least this is how it is supposed to be.
   */
  std::string getCommentHtml();
  
  /**
   *
   */
  virtual std::string getSomeText(eText whichText) = 0;
  
  /**
   * 
   */
  virtual void setKeywords(const char* nK) = 0;
  
  /**
   * 
   */
  virtual void setComment(const char* nC) = 0;
  
  /**
   * Previous item in the same group.
   */
  TreeItem* itemPrev;
  
  /**
   * Next item in the same group.
   * This is public to allow for fast group operations.
   */
  TreeItem* itemNext;   
  
  /**
   * Pointer to the TreeGroupBase to which I belong.
   * this is public to allow for a fast @ref treeDepth() (without 
   * recursive method calls).
   */
  TreeGroupBase* itemParent;
  
  /**
   * Returns the ID of the item this item points to (if it is a link).
   */
  virtual unsigned int getLinkID() { return(0); };
  
  /**
   * Returns the address of the item this item points to (if it is a link).
   */
  virtual TreeItem* getLinkPtr() { return((TreeItem*)0); };
  
  /**
   * 
   */
  virtual int loadFromXML_all(SimpleXMLTransfer*   source,
                              LinkInfo*            links);
  
  /**
   * 
   */
  virtual int saveToXML_all(SimpleXMLTransfer* dest, bool fConvLinks, int nVersion);
  
  /**
   * 
   */
  virtual enum eItemType getType() = 0;
  
  /**
   * 
   */
  virtual enum eItemLinkType getLinkType() = 0;
  
  /**
   * Jedes Element muss wissen, wie weit es eingerückt werden
   * muss. Dazu muss man die Anzahl der Schritte über @ref itemParent bis zum
   * Null-Element zählen. Es werden nicht viele gleichzeitig dargestellt, daher
   * kann man das jedesmal zählen. Alternativ könnte sich auch jede @ref TreeGroupBase
   * diese Entfernung merken.
   */
  int treeDepth();
  
  /**
   * Connect and/or count links. Returns 0 if everything is ok.
   */
  virtual int connectLinks(LinkInfo* links) { return(0); };
  
  /**
   * Find links in the comment and append the ID to which they point to the vector ids.
   */
  void appendCommentIDs(std::vector<unsigned int>& ids);
  
  /**
   * Increases/Decreases link count for targets found in the comment.
   * When creating/deleting an item, this must only be called for original comments!
   * 
   * Returns 0 if everything is fine.
   */
  int countCommentLinks(LinkInfo*    links,
                        bool         fIncr = true);
  
  /**
   * 
   */
  static const char* getTypeString(TreeItem::eItemType type);
  
  /**
   * 
   */
  unsigned int getID() { return(uID); };
  
  /**
   * 
   */
  void setID(unsigned int uNewID) { if (uID == 0) uID = uNewID; };
  
  /**
   * 
   */
  virtual void changeBasepath(std::string oldbp, std::string newbp) {};
  
  /**
   * 
   */
  virtual void kw_add(std::set<std::string> kwset, bool fEvenIfHidden = true);
  virtual void kw_replace(std::string from, std::string to, int mode);
  
  /**
   * 
   */
  virtual JWWDateTime getDT() = 0;
  
  /**
   * returns the number of missing files -- 0 if everything is fine
   */
  virtual unsigned int countMissingFiles(std::string basepath) { return(0); };
  
  /**
   * Adds keywords to the list (no dupes, alphabetically)
   */
  void kw_set(std::set<std::string>& kwset);
  
  /**
   * Adds keywords to the list (dupes, unchanged order)
   */
  void kw_list(std::vector<std::string>& kwlist);
  
  /**
   * Returns true if this item is hidden because of something
   */
  virtual bool isHidden();
  
  /**
   * Fills a set recursively (no duplicates).
   */
  virtual void RecursiveList(std::set<TreeItem*>& set);
  
  /**
   * File format conversion of comment string.
   */
  std::string ConvertComment(std::string in, int nNew);
  
protected:
  
  /**
   * My ID. Zero means 'no ID', I can not be linked.
   */
  unsigned int uID;
  
# if BASE_LNKCNT_IN_LIST == 0
  /**
   * How many links point to me?
   */
  unsigned int uLinkCnt;
# endif   
};
#endif
