// -*- mode: c; mode: fold -*-
//
/*
 * Copyright (C) 2005-2009, 2011  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "../config.h"
#include "gui.h"
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <string.h>
#include <iostream>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <FL/Fl.H>

# include <FL/fl_draw.H>
# include <FL/fl_ask.H>
# include <FL/Enumerations.H>
# include <FL/Fl_File_Chooser.H>
# include <FL/Fl_Tooltip.H>
# include <FL/Fl_Help_Dialog.H>

#ifdef USE_LIBGRAPHICSMAGICK
# include <magick/api.h>
#endif

#include "../shared/filetools.h"
#include "../shared/lib_conversions.h"
#include "../base/basetools.h"
#include "../base/export.h"
#include "picbutton.h"
#include "itempropdlg.h"
#include "workerdlg.h"
#include "searchdlg.h"
#include "kwreplacedlg.h"
#include "optsdlg.h"
#include "addpicsdlg.h"
#include "tips.h"
#include "exportdlg.h"
#include "exportmetadlg.h"
#include "albumoptdlg.h"
#include "misc.h"
#include "../base/treekwhandler.h"
#include "../shared/thread.h"

#include "../base/batch.h"
#include "../base/plugin.h"
#include "addkwdlg.h"

#include "popup.h"

/**
 * *********************** statische Daten ****************************************
 */

/**
 * Menü mit Recent Files, gilt für alle Fenster
 */
Fl_Menu_Item*  UserInterface::recentFiles_Menu = (Fl_Menu_Item*)0;

BlockTaskManager UserInterface::blockingTasks;
BackTaskDlg*     UserInterface::backTaskDlg = 0;

/**
 * Liste der Fenster
 */
std::vector<UserInterface*> UserInterface::windowList;

T_optionen*                 UserInterface::opts = (T_optionen*)0;

char**                      UserInterface::arguments;

bool                        UserInterface::fGimpHintShown = false;

Fl_Help_Dialog*             initHelpDlg;

/**
 * ********************************************************************************
 */


#define MENU_SAVE             3
#define MENU_RECENT           6
#define MENU_SHOWPICS        11
#define MENU_FULLSCREEN_PICS 12
#define MENU_FULLSCREEN      13
#define MENU_HELP            16
#define MENU_ABOUT           17

#include "callback.cpp"

// This is just the template for the main menu. Each instance of UserInterface
// creates its own local copy to provide a pointer to itself in the user data.
Fl_Menu_Item Main_Menu[] =
{
  {"&File",                 0,0,0,FL_SUBMENU},
  {"New",                   0,           cb_new},
  {"&Open...",              FL_CTRL+'o', cb_loadfile},
  {"&Save",                 FL_CTRL+'s', cb_savefile},
  {"Save as...",            0,           cb_savefileas},
  {"Album Settings...",     0,           cb_albset, 0, FL_MENU_DIVIDER},
  {"&Recent Files",         0, 0, 0, FL_SUBMENU_POINTER | FL_MENU_DIVIDER},
  {"Application &Options...",  0,           cb_options, 0, FL_MENU_DIVIDER},
  {"&Quit",                 FL_CTRL+'x', cb_ende},
  {0},
  {"&View",     0,0,0,FL_SUBMENU},
  {"Show Pics in TreeView", 0, (Fl_Callback*)cb_View, (void*)(UserInterface::SHOWPICS),        FL_MENU_TOGGLE},
  {"Fullscreen Pics",       0, (Fl_Callback*)cb_View, (void*)(UserInterface::FULLSCREEN_PICS), FL_MENU_TOGGLE},
  {"Fullscreen",            0, (Fl_Callback*)cb_View, (void*)(UserInterface::FULLSCREEN),      FL_MENU_TOGGLE},
  {0},
  {"&Help",     0,0,0,FL_SUBMENU},
  {"Help",                  0, (Fl_Callback*)cb_Help, (void*)0},
  {"About",                 0, (Fl_Callback*)cb_Help, (void*)1},
  {0},
  {0}
};

/**
 * workaround for thread-and-"xterm -e" problem: why do I receive SIGHUP?
 */
void signal_handler(int signum)
{
#ifndef WIN32
  std::cerr << "Received signal: " << signum << "  " << strsignal(signum) << "\n";
#endif
}

void setWaitCursor()
{
  fl_cursor(FL_CURSOR_WAIT);
  Fl::wait(0.1);
}
  

int execute_no_thread(void * (*start_routine)(void *), void * arg)
{
  int nRetCode;
  
  setWaitCursor();
  // work
  nRetCode = (int)(start_routine)(arg);
  // reset cursor
  fl_cursor(FL_CURSOR_DEFAULT);
  
  return(nRetCode);
}

void BlockingTaskPeriodic(void*)
{
  int nStateChange = UserInterface::blockingTasks.Step();
  switch (nStateChange)
  {
    case -1:
      UserInterface::backTaskDlg->w->hide();
      UserInterface::backTaskDlg->SetStatus(0, 0);
      break;
      
    case 1:
      UserInterface::backTaskDlg->w->show();
      break;
      
    default:
      break;
  }
  
  int a, b;
  if (!UserInterface::blockingTasks.Status(a, b))
  {
    UserInterface::backTaskDlg->SetStatus(a, b);
    UserInterface::blockingTasks.PauseUser(UserInterface::backTaskDlg->pPause->value() != 0);
    if (UserInterface::backTaskDlg->fCancel)
    {
      UserInterface::blockingTasks.Cancel();
      UserInterface::backTaskDlg->fCancel = false;
      UserInterface::backTaskDlg->pPause->value(0);
    }
  }
  
  Fl::repeat_timeout(0.2, BlockingTaskPeriodic);
}


int main(int argc, char **argv)                            
{
#ifndef UTF8_A  
  Fl::lock();
#endif

#ifndef WIN32
  // workaround for thread-and-"xterm -e" problem: why do I receive SIGHUP?
  signal (SIGHUP, signal_handler);
#endif
  
#ifdef USE_LIBGRAPHICSMAGICK
  InitializeMagick(*argv);
#endif    
    
  UserInterface::opts = new T_optionen((const char*)0);

  Fl::scheme(UserInterface::opts->UIStyle.c_str());

  UserInterface* GUI = new UserInterface();
  
  //
  std::string strFileToLoad = "";
  std::string strDirToLoad  = "";
  for (int n=1; n<argc; )
  {
    if (argv[n][0] == '-' && strlen(argv[n]) > 1)
    {
      switch (argv[n][1])
      {
        case 'd':
          strDirToLoad = argv[n+1];
          break;
          
        default:
          break;
      }
      n += 2;
    }
    else
    {
      if (n == argc-1)
        strFileToLoad = argv[n];
      n++;
    }
  }
  
  Fl_Double_Window* window = GUI->make_window(argc, argv);
  
  fl_register_images();

  // Von den Übergaben der Parameter an show hängt ab, ob die
  // im Windows eingestellten Farben übernommen werden. Wenn
  // einfach meine Parameter übergebe, geht das in die Hose,
  // weil er mit denen nichts anfangen kann. Also tue ich in
  // jedem Fall so, als wären keine Parameter (außer dem Aufruf selbst)
  // übergeben worden.
  UserInterface::arguments    = (char**)malloc(sizeof(char*)*1);
  UserInterface::arguments[0] = argv[0];
  window->show(1, UserInterface::arguments);

  if (!strFileToLoad.empty())
    GUI->loadFile(strFileToLoad.c_str());
  else if (!strDirToLoad.empty())
    GUI->loadDir(strDirToLoad.c_str(), 0, TreePicData::eEXIFFILEDATE);
  else if (UserInterface::opts->recentFilesCnt())
  {
    // load most recently used file
    GUI->loadFile(UserInterface::opts->recentFile(0).c_str());
  }

  if (UserInterface::opts->fShowInitialHelp)
  {
    std::string target;

    if (UserInterface::opts->nStartCnt < 0)
      target = "fa_initwarn.htm";
    else
    {
      target = StartupTips::getTipTarget(UserInterface::opts->nStartCnt);
    }
    
    initHelpDlg = new Fl_Help_Dialog();
    initHelpDlg->load(GUIMisc::getHelpPath("C", target).c_str());
    initHelpDlg->show();
  }
  else
  {
    initHelpDlg = (Fl_Help_Dialog*)0;
  }
  
  GUI->SetPermissions();

  UserInterface::backTaskDlg = new BackTaskDlg();
  UserInterface::backTaskDlg->w = UserInterface::backTaskDlg->make_window();
  UserInterface::backTaskDlg->w->hide();
  
  Fl::add_timeout(0.2, BlockingTaskPeriodic, 0);

  try
  {
    Fl::run();
  }
  catch (XMLException e)
  {
    std::cerr << "XMLException: " << e.what() << "\n";
  }

  while (UserInterface::windowList.size())
  {
    GUI = UserInterface::windowList[0];
    if (GUI != (UserInterface*)0)
    {
      delete GUI;
    }
  }
  
  delete window;
  
  if (PicBox::pPicCache)
    delete PicBox::pPicCache;

#ifdef USE_LIBGRAPHICSMAGICK
  DestroyMagick();
#endif      
  
  return 0;
}


Fl_Double_Window* UserInterface::make_window(int       argc,
                                             char*     argv[],
                                             TreeView* mysource) 
{
  window = new Fl_Double_Window(opts->nWinWidth, opts->nWinHeight);
#ifdef NOBORDER
  window->border(0);
#endif

  //
  Fl_Tooltip::size(opts->nFontSize);

  // Menü anlegen
  myMainMenu  = (Fl_Menu_Item*)malloc(sizeof(Main_Menu));
  for (unsigned int n=0; n<sizeof(Main_Menu)/sizeof(Fl_Menu_Item); n++)
  {
    myMainMenu[n] = Main_Menu[n];
    myMainMenu[n].labelsize(opts->nFontSize);

    switch (n)
    {
     case MENU_SHOWPICS:
     case MENU_FULLSCREEN_PICS:
     case MENU_FULLSCREEN:
     case MENU_HELP:
     case MENU_ABOUT:
      break;
     default:
      myMainMenu[n].user_data((void*)this);
      break;
    }
  }

  // Recent Files behandeln
  if (recentFiles_Menu == (Fl_Menu_Item*)0)
  {
    // ein leeres Menü anlegen
    recentFiles_Menu = (Fl_Menu_Item*)calloc(1, sizeof(Fl_Menu_Item));
    updateRecentFilesMenu();
  }

  myData = new TreeView(mysource);

  window->user_data((void*)(this));
  {
    int nHeight    = (opts->nFontSize * 16) >> 3;
    int nTreeWidth = opts->nTreeWidth;

    menuBar = new Fl_Menu_Bar(0, 0, window->w(), nHeight);
    menuBar->textsize (opts->nFontSize);

    tile     = new Fl_Tile(0, nHeight, window->w(), window->h()-nHeight, "tile");
    tile->labeltype(FL_NO_LABEL);

    if (nTreeWidth > tile->w() * 0.8)
      nTreeWidth = (int)(tile->w() * 0.8);

    browser  = new PicTreeBrowser(0, nHeight, nTreeWidth, tile->h(), "");
    browser->setData(myData);
    browser->setGUI(this);
    browser->labeltype(FL_NO_LABEL);
    browser->textsize(opts->nFontSize);
    browser->callback( (Fl_Callback*)cb_showItem, (void*)this );

    mainpane_pic = (PicBox*)0;

    mainpane_group = new PicGroup(nTreeWidth, nHeight, tile->w() - nTreeWidth, tile->h(), "");
    mainpane_group->labelsize(opts->nFontSize);
    mainpane_group->setData(myData);
    mainpane_group->setGUI(this);
    mainpane_group->user_data((void*)this);
    mainpane_group->end();

    // connect group and tree:
    mainpane_group->connectItemChange(browser);
    browser->connectItemChange(mainpane_group);

    tile->end();
    window->resizable(tile);
  }
  window->end();
  window->callback((Fl_Callback*)cb_ende, (void*)this);
  menuBar->menu(myMainMenu);
  
  // set options
  if (opts->pool->getInt("view.Fullscreen"))
    myMainMenu[MENU_FULLSCREEN_PICS].set();
  
  if (opts->pool->getInt("view.FullscreenWin", 0))
  {
    myMainMenu[MENU_FULLSCREEN].set();
    window->fullscreen();
  }
  
  if (opts->pool->getInt("view.PicsInTree"))
  {
    myMainMenu[MENU_SHOWPICS  ].set();
    browser->setShowPics(true);
  }  

  return window;
}


UserInterface::UserInterface()                             
{
  pDisplayReq = 0;
  windowList.push_back(this);
  myFilename = "";

  {
    std::cout << "Fenster: " << windowList.size() << " Stück\n";
    for (unsigned int n=0; n<windowList.size(); n++)
      std::cout << "  " << windowList[n] << "\n";
  }

  myData             = (TreeView*)0;
  popup              = (Fl_Menu_Button*)0;
  lastDisplayedItem  = (TreeItem*)0;
  lastDisplayedGroup = 0;  
  windowed           = 0;
  xmlpluginlist      = 0;
}


UserInterface::~UserInterface()                            
{
  for (unsigned int n=0; n<windowList.size(); n++)
  {
    if (windowList[n] == this)
    {
      // Mich selbst aus der Liste löschen
      windowList.erase(windowList.begin() + n);
      // Damit die Schleife beendet wird
      n = windowList.size();
    }
  }

  if (windowList.size() == 0 && initHelpDlg != (Fl_Help_Dialog*)0)
    initHelpDlg->hide();

  {
    std::cout << "Fenster: " << windowList.size() << " Stück\n";
    for (unsigned int n=0; n<windowList.size(); n++)
      std::cout << "  " << windowList[n] << "\n";
  }

  if (opts != (T_optionen*)0)
  {
    SimpleXMLTransfer* tmp;
    tmp = opts->pool->getChild("view");
    tmp->setAttributeOverwrite("Fullscreen",    myMainMenu[MENU_FULLSCREEN_PICS].value());
    tmp->setAttributeOverwrite("FullscreenWin", myMainMenu[MENU_FULLSCREEN].value());
    tmp->setAttributeOverwrite("PicsInTree",    myMainMenu[MENU_SHOWPICS  ].value());
    delete opts;
    opts = (T_optionen*)0;
  }

  if (myData != (TreeView*)0)
  {
    delete myData;
    myData = (TreeView*)0;
  }
  
  if (mainpane_pic != 0)
    delete mainpane_pic;

  while (groupviewdata.size())
  {
    TGroupViewData gvd;
    
    gvd = groupviewdata[groupviewdata.size()-1];
    groupviewdata.pop_back();
    
    if (gvd.thumbnail != (Fl_Shared_Image*)0)
    {
      gvd.thumbnail->uncache();
      gvd.thumbnail->release();
    }
    free(gvd.tooltip);
  }
  
  free(myMainMenu);
  
  if (xmlpluginlist)
    delete xmlpluginlist;  
  if (popup != 0)
    delete popup;  
}


int  UserInterface::loadFile(const char* filename)         
{
  int         nRetCode;

  myFilename = filename;

  if (myData != (TreeView*)0)
    delete myData;
  myData = new TreeView((TreeView*)0);
  nRetCode = myData->loadFromFile(filename);

  if (nRetCode == 0)
  {    
    opts->addRecentFile(filename);
    updateRecentFilesMenu();
  }
  
  if ((nRetCode & TreeView::eLE_LINK) != 0)
    fl_alert("Error in project file: link error (see console)");
  if ((nRetCode & TreeView::eLE_BASEPATH) != 0)
    fl_alert("Error: basepath");
  if ((nRetCode & TreeView::eLE_XML) != 0)
    fl_alert("Error in project file: XML structure (see console)");
  if ((nRetCode & TreeView::eLE_VERSION) != 0)
    fl_alert("Error: project file has been created with a more recent\nversion of fotoalbum; I can't handle that. Please update!");
  if ((nRetCode & TreeView::eLE_NOFILE) != 0)
  {
    fl_alert("Error: no such file");
    // remove that file from recent files menu
    opts->removeRecentFile(filename);
    updateRecentFilesMenu();
  }

  browser->setData(myData);
  if (mainpane_group != (PicGroup*)0)
    mainpane_group->setData(myData);
  browser->redraw();

  setWindowTitle(filename);

  return(0);
}


void UserInterface::setWindowTitle(const char* filename)   
{
  strWindowTitle = TITLE_START;

  if (filename != (const char*)0)
    strWindowTitle += filename;
  
  // valgrind output indicates that this string is not copied, so I have to maintain
  // that buffer:
  window->label(strWindowTitle.c_str());
}


void UserInterface::updateRecentFilesMenu()                
{
  // Das Menü neu machen
  // alte Einträge freigeben (Strings)
  int n = 0;
  while (recentFiles_Menu[n].text != (const char*)0)
  {
    printf("Freigeben von %i\n", n);

    free((void*)recentFiles_Menu[n].text);
    recentFiles_Menu[n].text = (const char*)0;
    n++;
  }

  // Neue Menüs anlegen
  if (opts->recentFilesCnt() >= 0)
  {
    std::string  rftext;
    int          nSize = opts->recentFilesCnt();

    // Neue Größe erzeugen
    recentFiles_Menu = (Fl_Menu_Item*)realloc(recentFiles_Menu, sizeof(Fl_Menu_Item)*(nSize+1));

    for (n=0; n<nSize; n++)
    {
      // Text
      rftext = opts->recentFile(n);

      // todo: schlauer machen!
      while (rftext.length() > 50)
        rftext = rftext.substr(1);

      recentFiles_Menu[n].text = (char*)malloc(sizeof(char) * (rftext.length()+1));
      strcpy((char*)recentFiles_Menu[n].text, rftext.c_str());
      //
      recentFiles_Menu[n].user_data_ = (void*)n;
      //
      recentFiles_Menu[n].callback_ = (Fl_Callback*)cb_loadrecentfile;
      recentFiles_Menu[n].labelsize_  = opts->nFontSize;
      // Der Rest ist null
      recentFiles_Menu[n].shortcut_   = 0;
      recentFiles_Menu[n].flags       = 0;
      recentFiles_Menu[n].labeltype_  = 0;
      recentFiles_Menu[n].labelfont_  = 0;
      recentFiles_Menu[n].labelcolor_ = 0;
    }
    // Null-Element anhängen
    recentFiles_Menu[nSize].text = (const char*)0;
  }

  // Jedes Fenster muss jetzt den neuen Zeiger verwenden
  for (unsigned int n=0; n<windowList.size(); n++)
    windowList[n]->updateRecentFilesMenuPtr();
}


void UserInterface::updateRecentFilesMenuPtr()             
{
  myMainMenu[MENU_RECENT].user_data((void*)recentFiles_Menu);
}


void UserInterface::enterFullscreen()                      
{
  Fl_Widget* mainpane;

  if (mainpane_group == 0)
    mainpane = mainpane_pic;
  else
    mainpane = mainpane_group;

  // save old state
  windowed = new TWindowedData();
  windowed->win_x = window->x();
  windowed->win_y = window->y();
  windowed->win_w = window->w();
  windowed->win_h = window->h();
  windowed->pane_x = mainpane->x();
  windowed->pane_y = mainpane->y();
  windowed->pane_w = mainpane->w();
  windowed->pane_h = mainpane->h();
  // switch to fullscreen mode
#if (1 == 0)
  printf("%i %i %i %i\n", Fl::x(), Fl::y(), Fl::w(), Fl::h());
  window->resize(Fl::x(), Fl::y(), Fl::w(), Fl::h());
#endif
  window->fullscreen();
  menuBar->hide();
  browser->hide();
}

int  UserInterface::displayItem(TreeItem* item, bool fAllowFullscreen)
{
  if (pDisplayReq == 0)
  {
    pDisplayReq = item;
    fDispReqFullscreen = fAllowFullscreen;
    
    Fl::add_idle(cb_idle, (void*)(this));      
  }
  return(0);
}

void UserInterface::setPicFullscreen()                     
{
  std::cout << "Pics Fullscreen = " << myMainMenu[MENU_FULLSCREEN_PICS].value() << "\n";

  if (myMainMenu[MENU_FULLSCREEN_PICS].value() &&
      mainpane_group == 0)
  {
    int x;
    int y;
    int w;
    int h;

    enterFullscreen();

    x = window->x();
    y = window->y();
    w = window->w();
    h = window->h();

    mainpane_pic->resize(x, y, w, h);
  }
}

void UserInterface::setFullscreen()                     
{
  if (myMainMenu[MENU_FULLSCREEN].value())
  {
    window->fullscreen();    
  }
  else
  {
    // which size and offset will work on every system?
    int x = (2*window->w())/3;
    int y = (2*window->h())/3;
    window->fullscreen_off(x/4, y/4, x, y);
  }
}


void UserInterface::selectCurrentItem()                    
{
  TreeItem* item;

  if (mainpane_group == 0)
  {
    item = mainpane_pic->getPic();
    if (item != (TreeItem*)0 && browser->showPics() == false)
      item = item->itemParent;
  }
  else
    item = browser->itemSelected;

  if (item != (TreeItem*)0)
  {
    if (myData->makeVisible(item))
    {
      browser->redraw();
    }

    browser->display(item);
    browser->select_only(item);
  }
}


bool UserInterface::hasData()                              
{
  return((myData != (TreeView*)0) && myData->hasData() && myData->getSource() == (TreeView*)0);
}


std::string UserInterface::stringMaxLen(std::string  in,
                                        unsigned int mlen) 
{
  if (in.length() > mlen)
  {
    in = in.substr(0, mlen-3) + "...";
  }
  return(in);
}


void UserInterface::followLink(std::string linktext)       
{
  std::cout << "void UserInterface::followLink(std::string linktext = " << linktext << ")\n";

  TreeItem* item = myData->resolveInternalLink(linktext);

  if (item != (TreeItem*)0)
  {
    displayItem(item);
  }
  else
  {
    std::string path = myData->resolveExternalLink(linktext, lastDisplayedItem);
    std::cout << "display file: " << path << "\n";
    Fl_Help_Dialog* dlg = new Fl_Help_Dialog();
    dlg->load(path.c_str());
    dlg->show();
  }
}


int  UserInterface::showContextMenu(int       m_x,
                                    int       m_y,
                                    TreeItem* item)        
{
  unsigned int           uItemCnt;
  TreeItem::eItemType    type;

  Fl_Tooltip::enable(0);

  std::cout << "context menu at " << m_x << "," << m_y << "\n";

  if (myData->countMarkedItems() == 0 ||
      (myData->countMarkedItems() != 0 && !item->fMarked))
  {
    // Menu for a single item
    uItemCnt            = 1;
    type                = item->getType();
    lastPopupSingleItem = item;
  }
  else
  {
    // Menu for a multiple items
    uItemCnt            = myData->countMarkedItems();
    type                = myData->markedItemsType();
    if (uItemCnt == 1)
    {
      lastPopupSingleItem = myData->getFirstMarkedItem();
    }
    else
      lastPopupSingleItem = (TreeItem*)0;
  }

  std::cout << "Context menu: " << uItemCnt << " items, type: " << type << "\n";
  
  if (popup != 0)
    delete popup;  
  popup = new Fl_Menu_Button(m_x, m_y, 10, 0, "popup");
  popup->type(0);

  // Info entry
  {
    std::string info;
    if (uItemCnt == 1)
    {
      info = "(One item of type ";
      info += TreeItem::getTypeString(type);
      info += ")";
    }
    else
    {
      info = "(" + itoStr(uItemCnt, ' ', 1) + " items of type " + TreeItem::getTypeString(type) + ")";
    }
    addToPopup(true, info.c_str(), PA_NONE);
  }

  // Real entries
  if (uItemCnt == 1 && IsPermitted(PERM_EditMeta))
    addToPopup(true, "Properties...",               PA_PROPERTIES);

  if (uItemCnt == 1 && type == TreeItem::GROUP)
  {
    if (IsPermitted(PERM_AddGroups))
      addToPopup(false, "Add/Group...",               PA_ADDGROUP);
    if (myData->isAddDelAllowed() && IsPermitted(PERM_AddFiles))
    {
      addToPopup(false, "Add/Files...",             PA_ADDPICS);
      addToPopup(false, "Add/File...",              PA_ADDPIC);
    }
    if (IsPermitted(PERM_MoveItems))
    {
      addToPopup(false, "Order/chronologically",      PA_ORDER_C);
      addToPopup(false, "Order/by filename",          PA_ORDER_P);
      addToPopup(false, "Order/by filesize",          PA_ORDER_S);
    }
    addToPopup(true, "All files available?",        PA_CHECK_FILES);
    addToPopup(false, "View chronologically",       PA_VIEW_CHRONO);
    addToPopup(false, "Search or mark...",          PA_SEARCH);
    addToPopup(true, "Show items without text",     PA_SEARCH_WO_TEXT);
    if (IsPermitted(PERM_EditMeta))
     addToPopup(true, "Edit keywords...",            PA_KWREPLACE);
  }

  if (type == TreeItem::GROUP && myData->isAddDelAllowed() && IsPermitted(PERM_CreateThumbs))
    addToPopup(true,  "Create Thumbnails",          PA_THUMB_CREATE);

  if (type == TreeItem::PIC)
  {
    if (myData->isAddDelAllowed() && IsPermitted(PERM_EditFiles))
    {
      addToPopup(false, "Rotate counterclockwise",    PA_ROT_CCW);
      addToPopup(false, "Rotate clockwise",           PA_ROT_CW);
      addToPopup(false, "Open with GIMP",             PA_EDIT_GIMP);
    }
    if (IsPermitted(PERM_EditMeta))
      addToPopup(false, "Shift time...",              PA_TIMESHIFT);
    if (IsPermitted(PERM_AppendMeta))
      addToPopup(true,  "Add keyword...",             PA_ADD_KEYWORD);
    
    // Insert Plugins menu
    {
      if (xmlpluginlist)
        delete xmlpluginlist;
      
      std::string loadErrorString = Plugin::getPluginList(xmlpluginlist);
      
      if (loadErrorString.length())
        fl_alert(loadErrorString.c_str());
      else if (xmlpluginlist)
      {
        SimpleXMLTransfer* g;
        unsigned int       size;

        try
        {
          g = xmlpluginlist->getChild("item");
          size = g->getChildCount();
          
          if (size)
          {
            SimpleXMLTransfer*     i;
            std::string            label;            
            std::vector<TreeItem*> list;
            
            if (lastPopupSingleItem != (TreeItem*)0)
              list.push_back(lastPopupSingleItem);
            else
              myData->getMarkedItems(list);
            
            for (unsigned int cnt=0; cnt<size; cnt++)
            {
              i = g->getChildAt(cnt);
              
              if (Plugin::match(list, i) && 
                  (i->attributeAsInt("readonly", 1)==1 || IsPermitted(PERM_EditFiles)))
              {
                label = "Plugins/" + i->attribute("descr");              
                addToPopup((cnt == size-1),  label.c_str(),   PA_PLUGIN_BASE + cnt);
              }                            
            }                    
          }
        }
        catch (XMLException e)
        {
          std::string msg = "Error in plugins.xml: ";
          msg += e.what();
          fl_alert(msg.c_str());
        }  
      }
    }
  }
  
  if (uItemCnt == 1 && myData->countMarkedItems() > 0 && !item->fMarked)
  {
    if (item->getType() == TreeItem::GROUP)
    {
      if (IsPermitted(PERM_MoveItems))
        addToPopup(false, "Move marked into this group",      PA_MOVE_INTO);
      if (IsPermitted(PERM_AddLinks))
        addToPopup(false, "Link marked into this group",      PA_LINK_INTO);
    }

    if (IsPermitted(PERM_MoveItems))
      addToPopup(false, "Move marked after this item",        PA_MOVE_AFTER);
    if (IsPermitted(PERM_AddLinks))
      addToPopup(true,  "Link marked after this item",        PA_LINK_AFTER);
  }

  addToPopup(true,  "Import, Export/Export...",                 PA_EXPORT);
  addToPopup(true,  "Import, Export/Export metadata only...",   PA_EXPORT_META);
  if (IsPermitted(PERM_EditMeta))
    addToPopup(true,  "Import, Export/Import metadata",           PA_IMPORT_META);
  
  {
    bool fErlaubt = IsPermitted(PERM_DelItems);
    if (!fErlaubt && IsPermitted(PERM_DelEmptyGroups) && uItemCnt == 1 && type == TreeItem::GROUP)
    {
      // prüfen, ob die Gruppe vielleicht leer ist
      TreeGroupBase* group = (TreeGroupBase*)(lastPopupSingleItem);
      fErlaubt = (group->itemFirstChild == 0);        
    }
    if (fErlaubt)
      addToPopup(false, "Remove from album",            PA_REMOVE);
  }
  
  if (IsPermitted(PERM_DelFiles))
    addToPopup(false, "Remove (incl. picture files)", PA_WIPE);

  popup->popup();

  return(0);
}


void UserInterface::addToPopup(bool        fSep,
                               const char* descr,
                               int         nAction)        
{
  std::string Descr;
  if (fSep)
    Descr = "_";
  else
    Descr = "";
  Descr.append(descr);
  popup->add(Descr.c_str(), 0, (Fl_Callback*)cb_popup, (void*)nAction);
}

std::string UserInterface::GetSourcepath(TreeGroupBase* group)
{
  return(myData->GetAddPath());
}


void UserInterface::popupAction(int nAction)               
{
  std::cout << "Popup action " << nAction << "\n";

  // Many actions can be applied to one or more items.
  // Here I put some into a list. This is not needed for some
  // actions, but it reduces the amount of code in this method
  // (478 lines to 429 lines).
  std::vector<TreeItem*> list;

  if (lastPopupSingleItem != (TreeItem*)0)
    list.push_back(lastPopupSingleItem);
  else
    myData->getMarkedItems(list);

  switch (nAction)
  {
   case PA_NONE:
    break;

   case PA_CHECK_FILES:
    {
      setWaitCursor();
      unsigned int uMF = ((TreeGroupBase*)lastPopupSingleItem)->countMissingFiles(myData->getBasepath());
      fl_cursor(FL_CURSOR_DEFAULT);
      
      if (uMF)
        fl_alert("At least one file is missing! See console...");
      else
        fl_message("All files found.");
    }
    break;

   case PA_ADDGROUP:
    {
      const char* newnameptr = fl_input("Name of new group", "new group");
      if (newnameptr != (const char*)0)
      {
        std::string newname = newnameptr;
        if (newname.length() > 0)
        {
          ((TreeGroupBase*)lastPopupSingleItem)->addNewGroup(newname.c_str(), TreeItem::DATA);
          browser->redraw();
        }
      }
    }
    break;

   case PA_ORDER_C:
    {
      setWaitCursor();
      ((TreeGroupBase*)lastPopupSingleItem)->sortChildrenChrono();
      fl_cursor(FL_CURSOR_DEFAULT);
    }
    break;

   case PA_ORDER_P:
    {
      ((TreeGroupBase*)lastPopupSingleItem)->sortChildrenPath();
    }
    break;

   case PA_ORDER_S:
    {
      ((TreeGroupBase*)lastPopupSingleItem)->sortChildrenSize(myData->getBasepath());
    }
    break;

   case PA_PROPERTIES:
    {
      // may be triggered by keyboard-shortcut, therefore permissions have to be checked:
      if (IsPermitted(PERM_EditMeta))
      {
        ItemPropertyDlg*  dlg = new ItemPropertyDlg(lastPopupSingleItem, myData);
        dlg->make_window();
      }
    }
    break;

   case PA_LINK_INTO:
    {
      myData->linkInto_marked((TreeGroupBase*)lastPopupSingleItem);
      redrawViews();
    }
    break;

   case PA_MOVE_INTO:
    {
      myData->moveInto_marked((TreeGroupBase*)lastPopupSingleItem);
      redrawViews();
    }
    break;

   case PA_MOVE_AFTER:
    {
      myData->moveAfter_marked(lastPopupSingleItem);
      redrawViews();
    }
    break;

   case PA_ADDPICS:
    {
      char*        picdir;

      {
        std::string bp = GetSourcepath((TreeGroupBase*)lastPopupSingleItem);

        picdir = fl_dir_chooser("Insert files from directory", bp.c_str());
      }

      if (picdir != (char*)0)
      {
        TreePicData::eDateTimeSource dtmethod = TreePicData::eEXIFFILEDATE;
        
        if (false)
        {
          AddPicsDlg* dlg = new AddPicsDlg();
          dtmethod = dlg->showDlg();
          delete dlg;
        }

        if (dtmethod != TreePicData::eABORT)
          loadDir(picdir, (TreeGroupBase*)lastPopupSingleItem, dtmethod);
      }
    }
    break;
    
   case PA_ADDPIC:
    {
      char*        picdir;

      {
        std::string bp = GetSourcepath((TreeGroupBase*)lastPopupSingleItem);

        picdir = fl_file_chooser("Insert file", NULL, bp.c_str());
      }

      if (picdir != (char*)0 && strlen(picdir) > 0)
      {
        TreePicData::eDateTimeSource dtmethod = TreePicData::eEXIFFILEDATE;

        if (false)
        {
          AddPicsDlg* dlg = new AddPicsDlg();
          dtmethod = dlg->showDlg();
          delete dlg;
        }

        if (dtmethod != TreePicData::eABORT)
          myData->addFile(picdir,
                          (TreeGroupBase*)lastPopupSingleItem,
                          dtmethod);
      }
    }
    break;
      
    case PA_KWREPLACE:
      {
        KwReplaceDlg* dlg = new KwReplaceDlg((TreeGroupBase*)lastPopupSingleItem, myData->root);
        dlg->showDlg();
        delete dlg;
      }
      break;
      
    case PA_SEARCH:
      {
        std::set<std::string> kwset;        
        ((TreeGroupBase*)lastPopupSingleItem)->kw_setGroup(kwset);
        
        SearchDlg* dlg = new SearchDlg(kwset);
        int nRetCode = dlg->showDlg();
        
        if (nRetCode != 0)
        {
          SearchData sd;
          dlg->setData(&sd);
          
          if (sd.getNrOfWords())
          {
            switch (nRetCode)
            {
              case 2:
                {
                  UserInterface*    newGUI = new UserInterface();
                  Fl_Double_Window* window = newGUI->make_window(1, UserInterface::arguments, myData);
                  
                  newGUI->AfterCreation();
                  // bisherige Markierungen aufheben
                  myData->resetMarkItem((TreeItem*)0);
                  //
                  newGUI->newViewBySearch((TreeGroupBase*)lastPopupSingleItem, sd);
                  newGUI->setWindowTitle("(search result)");
                  window->show(1, UserInterface::arguments);
                }
                break;
                
              default:
                {
                  myData->markItems((TreeGroupBase*)lastPopupSingleItem, sd);
                }
            }            
          }
        }
        delete dlg;
      }
      break;
      
    case PA_SEARCH_WO_TEXT:
    {
      setWaitCursor();
      
      UserInterface*    newGUI = new UserInterface();
      Fl_Double_Window* window = newGUI->make_window(1, UserInterface::arguments, myData);

      newGUI->AfterCreation();
      myData->resetMarkItem((TreeItem*)0);
      newGUI->newViewWithoutText((TreeGroupBase*)lastPopupSingleItem);
      newGUI->setWindowTitle("(items without text)");
      window->show(1, UserInterface::arguments);
      
      fl_cursor(FL_CURSOR_DEFAULT);
    }
    break;

  case PA_VIEW_CHRONO:
    {
      setWaitCursor();
      
      UserInterface*    newGUI = new UserInterface();
      Fl_Double_Window* window = newGUI->make_window(1, UserInterface::arguments, myData);
      
      newGUI->AfterCreation();
      newGUI->newViewByDate((TreeGroupBase*)lastPopupSingleItem);
      newGUI->setWindowTitle("(chronological view)");
      window->show(1, UserInterface::arguments);
      
      fl_cursor(FL_CURSOR_DEFAULT);
    }
    break;

   case PA_TIMESHIFT:
    {
      const char* newnameptr = fl_input("Seconds to shift (positive/negative)", "-3600");
      if (newnameptr != (const char*)0)
      {
        std::string newname = newnameptr;
        if (newname.length() > 0)
        {
          int nShift = atoi(newname.c_str());

          BatchProc::timeshift(list, nShift);
        }
      }
    }
    break;

    case PA_ADD_KEYWORD:
      // may be triggered by keyboard-shortcut, therefore permissions have to be checked:
      if (IsPermitted(PERM_AppendMeta))
      {
        std::set<std::string> kwset;        
        myData->root->kw_setGroup(kwset);
        AddKwDlg* dlg = new AddKwDlg(kwset);
        if (dlg->showDlg())
        {
          const char* kwstringptr = dlg->ti_words->value();
          if (kwstringptr != (const char*)0)
          {
            std::set<std::string> kwset = BaseTools::KeywordList_Set(kwstringptr);

            if (!kwset.empty())
              BatchProc::kw_add(list, kwset);
          }
        }
        delete dlg;
      }
      break;

    case PA_ROT_CCW:
      for (std::vector<TreeItem*>::iterator it = list.begin(); it!=list.end(); it++)
      {
        TreePicBase* pPic = dynamic_cast<TreePicBase*>(*it);
        if (pPic && pPic->getType() == TreeItem::PIC)
          blockingTasks.Add(new BlockingTask_Rotate(myData->getBasepath() + pPic->getPath(), 90));
      }
      break;

    case PA_ROT_CW:
      for (std::vector<TreeItem*>::iterator it = list.begin(); it!=list.end(); it++)
      {
        TreePicBase* pPic = dynamic_cast<TreePicBase*>(*it);
        if (pPic && pPic->getType() == TreeItem::PIC)
          blockingTasks.Add(new BlockingTask_Rotate(myData->getBasepath() + pPic->getPath(), -90));
      }
      break;

   case PA_EDIT_GIMP:
    {
      if (!fGimpHintShown)
      {
        fGimpHintShown = true;
        // Last time I used it, it worked, although gimp had not been running.
        // I remember that it did not work someday...
        fl_message("This may only work if Gimp is already running.");
      }
      setWaitCursor();
      BatchProc::editGimp(list, myData->getBasepath());
      fl_cursor(FL_CURSOR_DEFAULT);
    }
    break;

   case PA_EXPORT: 
    {
      ExportDlg* dlg = new ExportDlg(opts, list[0]->getType() == TreeItem::GROUP);
      if (dlg->showDlg())
      {
        Export ex;
        ex.destdir = opts->pool->getString("export.dir");
        if (!ex.destdir.empty())
        {
          ex.setOpts(opts->pool);
          
          ex.data = myData;
          
          // fixed values because files are only created, so there is nothing to merge/overwrite
          ex.fKwInherit        = true;
          ex.fKwMerge          = false;
          ex.fCommentOverwrite = true;
          
          ex.run(list);
          if (ex.error.length())
          {
            fl_alert(ex.error.c_str());
          }
          else
          {
            opts->pool->setAttributeOverwrite("export.dir", ex.destdir);
            
            WorkThumbCreate work;
            WorkerData      wd;
            int             nRetCode;
            
            // --- Part 1/4: shrink list --------------------------
            {
              SimpleXMLTransfer* shr = opts->getShrink();
              work.nMaxW  = shr->attributeAsInt("w", 800);
              work.nMaxH  = shr->attributeAsInt("h", 600);
              work.flQual = opts->pool->getDouble("shrink.quality");
              if (opts->pool->getInt("export.fast", 0) != 0)
                work.nType = 2;
              else
                work.nType = 0;
              for (unsigned int cnt=0; cnt<ex.shrink.size(); cnt++)
              {
                work.filelist.push_back(ex.shrink[cnt]);
              }
            }
            
            wd.init(work.filelist.size()/2, (void*)&work);
            
            nRetCode = SimpleThread::Create(createThumbs, (void*)&wd);
            if (nRetCode)
            {
              std::cerr << "error creating thread\n";
              std::cerr.flush();
              nRetCode = -1;
            }
            else
            {
              blockingTasks.Pause(true);
              WorkerDlg* wdlg = new WorkerDlg(&wd);
              wdlg->showDlg();
              while (wd.nRun) { Fl::wait(0.2); };
              nRetCode = wd.nRetCode;
              delete wdlg;
              blockingTasks.Pause(false);
            }
            if (nRetCode)
            {
              fl_alert("Error");
            }
            
            // --- Part 2/4: thumb list --------------------------
            {
              work.nMaxW  = opts->pool->getDouble("export.thumbsize");
              work.nMaxH  = work.nMaxW;
              work.flQual = opts->pool->getDouble("export.thumbqual");
              if (opts->pool->getInt("export.fast", 0) != 0)
                work.nType = 2;
              else
                work.nType = 0;
              work.filelist.clear();
              for (unsigned int cnt=0; cnt<ex.vThumbs.size(); cnt++)
              {
                work.filelist.push_back(ex.vThumbs[cnt]);
              }
            }
            
            wd.init(work.filelist.size()/2, (void*)&work);
            
            nRetCode = SimpleThread::Create(createThumbs, (void*)&wd);
            if (nRetCode)
            {
              std::cerr << "error creating thread\n";
              std::cerr.flush();
              nRetCode = -1;
            }
            else
            {
              blockingTasks.Pause(true);
              WorkerDlg* wdlg = new WorkerDlg(&wd);
              wdlg->showDlg();
              while (wd.nRun) { Fl::wait(0.2); };
              nRetCode = wd.nRetCode;
              delete wdlg;
              blockingTasks.Pause(false);
            }
            if (nRetCode)
            {
              fl_alert("Error");
            }
            
            
            
            // --- Part 3/4: copy list ----------------------------
            work.filelist.clear();
            {
              for (unsigned int cnt=0; cnt<ex.copy.size(); cnt++)
              {
                work.filelist.push_back(ex.copy[cnt]);
              }
            }
            
            wd.init(work.filelist.size()/2, (void*)&work);
            
            nRetCode = SimpleThread::Create(copyFiles, (void*)&wd);
            if (nRetCode)
            {
              std::cerr << "error creating thread\n";
              std::cerr.flush();
              nRetCode = -1;
            }
            else
            {
              blockingTasks.Pause(true);
              WorkerDlg* wdlg = new WorkerDlg(&wd);
              wdlg->showDlg();
              while (wd.nRun) { Fl::wait(0.2); };
              nRetCode = wd.nRetCode;
              delete wdlg;
              blockingTasks.Pause(false);
            }
            if (nRetCode)
            {
              fl_alert("Error");
            }
            
            
            // --- Part 4/4: annotate list ------------------------
            setWaitCursor();
            ex.run_annotate();
            fl_cursor(FL_CURSOR_DEFAULT);              
          }
        }
      }
      delete dlg;
    }        
    break;

   case PA_EXPORT_META:
    {
      ExportMetaDlg* dlg = new ExportMetaDlg();
      if (dlg->showDlg())
      {
        Export ex;
        
        ex.type        = Export::eMetaOnly;
        ex.structure   = Export::eFileSys;
        ex.fRecursive  = true;
        ex.fThumbs     = false;
        ex.fFollowLinks= true;
        ex.fDatafile   = false;
        ex.fAddToJpeg  = true;
        ex.fPrefix     = false;
        ex.fHtml       = false;          
        
        ex.fKwInherit        = (dlg->sw_kwinherit   ->value() != 0);
        ex.fKwMerge          = (dlg->rb_kwmerge     ->value() != 0);
        ex.fCommentOverwrite = (dlg->sw_comoverwrite->value() != 0);
                
        ex.destdir     = myData->getBasepath();
        ex.data        = myData;
            
        ex.run(list);
        
        if (ex.error.length())
        {
          fl_alert(ex.error.c_str());
        }
        else
        {
          // --- Part 1/1: annotate list ------------------------
          setWaitCursor();
          ex.run_annotate();
          fl_cursor(FL_CURSOR_DEFAULT);              
        }
      }
      delete dlg;
    }        
    break;
      
    case PA_IMPORT_META:
      {
        std::set<TreeItem*> set;
        std::set<TreeItem*>::iterator it;
        TreeView::RecursiveList(list, set);
        it = set.begin();
        while (it != set.end())
        {
          TreePicData* p = dynamic_cast<TreePicData*>(*it);
          if (p)
            p->ImportMetadata(myData->getBasepath());
          it++;
        }
      }
      break;

   case PA_REMOVE:
   case PA_WIPE:
    {        
      bool fDelFiles = (nAction == PA_WIPE);
      int  nC;

      if (fDelFiles)
        nC = fl_choice("Really delete? Image files will be deleted, too!", 0, "Cancel", "Delete");
      else
        nC = fl_choice("Really remove from album?", 0, "Cancel", "Remove");

      if (nC == 2)
      {
        int nRetCode;

        if (lastPopupSingleItem != (TreeItem*)0)
        {
          if (lastDisplayedItem == lastPopupSingleItem)
            lastDisplayedItem = (TreeItem*)0;
        }
        else
        {
          if (lastDisplayedItem->fMarked)
            lastDisplayedItem = (TreeItem*)0;
        }

        nRetCode = BatchProc::deleteItems(list, myData, fDelFiles);

        if (nRetCode != 0)
        {
          if (nRetCode < 4)
            fl_alert("Error: Maybe at least one item not deleted because\n"
                     " -it is being pointed to\n"
                     " -it is a group which is not empty\n");

          if ((nRetCode & eLiErrUnderflow) != 0)
            fl_alert("Critical internal error: link underflow");
          if ((nRetCode & eLiErrIDNotInList) != 0)
            fl_alert("Critical internal error: link ID could not be found");
        }
        redrawViews();
      }
    }
    break;

    default:
      if (nAction >= PA_PLUGIN_BASE)
      {
        std::string strErr;
        SimpleXMLTransfer* plugin = xmlpluginlist->getChild("item")->getChildAt(nAction - PA_PLUGIN_BASE);
        if (plugin->attributeAsInt("blocking", 0))
        {
          std::list<std::string> vCmds;
          strErr = Plugin::getCmds(list, plugin, myData, vCmds);
          if (strErr.empty())
            for (std::list<std::string>::iterator it = vCmds.begin(); it!=vCmds.end(); it++)
              blockingTasks.Add(new BlockingTask_System(*it));
        }
        else
          strErr = Plugin::run(list, plugin, myData);
        
        if (strErr.length())
          fl_alert(strErr.c_str());        
      }
      else  
        fl_message("Not implemented yet");
      break;
  }
}


int  UserInterface::save(bool fNewName)                    
{
  int nRetCode = 0;

  if (myData->getProjectfilePath().length() == 0)
    fNewName = true;

  if (fNewName)
  {
    char*        newfile;

    newfile = fl_file_chooser("Save as...", "*.pg.xml", "");

    if (newfile != (char*)0)
    {
      std::string filename = newfile;

      if (filename.find(".pg.xml") != filename.length()-7)
      {
        filename += ".pg.xml";
      }

      opts->addRecentFile(filename.c_str());
      updateRecentFilesMenu();

      nRetCode = myData->saveAsNewProject(filename);
      setWindowTitle(filename.c_str());

      if (nRetCode)
      {
        fl_alert("Error saving to new file");
      }
    }
  }
  else
  {
    nRetCode = myData->saveToProjectFile();
    if (nRetCode)
    {
      fl_alert("Error saving to file");
    }
  }
  
  SetPermissions();
  
  return(nRetCode);
}


int  UserInterface::askBeforeClosing()                     
{
  bool fUnsavedData = true;

  if (hasData())
  {
    setWaitCursor();
    
    std::cout << "hasData() == true\n";
    if (myData->getProjectfilePath().length() != 0)
    {
      std::cout << "projectfilePath\n";
      if (myData->dataSaved())
        fUnsavedData = false;
    }
    
    fl_cursor(FL_CURSOR_DEFAULT);
  }
  else
  {
    std::cout << "hasData() == false\n";
    fUnsavedData = false;
  }

  if (fUnsavedData)
  {
    if (fl_choice("There is unsaved data. Exit anyway?", NULL, "Cancel", "Exit") == 2)
      fUnsavedData = false;
  }

  if (fUnsavedData)
    return(-1);
  else
    return(1);
}


void UserInterface::newViewByDate(TreeGroupBase* group)    
{
  myData->newViewByDate(group);
  displayItem(myData->root);
}


void UserInterface::newViewBySearch(TreeGroupBase* group,
                                    SearchData     sd)     
{
  myData->newViewBySearch(group, sd);
  displayItem(myData->root);
}


void UserInterface::newViewWithoutText(TreeGroupBase* group)  
{
  myData->newViewWithoutText(group);
  displayItem(myData->root);
}


void UserInterface::redrawViews()                             
{
  browser->redraw();
  if (mainpane_group != (PicGroup*)0 && lastDisplayedItem != (TreeItem*)0)
  {
    displayItem(lastDisplayedItem);
  }
}


void UserInterface::saveOptions()                             
{
  opts->nWinWidth  = window->w();
  opts->nWinHeight = window->h();
  opts->nTreeWidth = browser->w();
}


void UserInterface::optionsdlg()                              
{
  OptsDlg* dlg = new OptsDlg(opts);
  dlg->showDlg();
  delete dlg;
}


void UserInterface::setLastDisplayedAsLastPopupSingle()       
{
  lastPopupSingleItem = lastDisplayedItem;
}


void UserInterface::toggleMarkCurrentPic() 
{
  // toggle data
  myData->toggleMarkItem(mainpane_pic->getPic());
  // make it visible in the tree
  browser->itemChanged(mainpane_pic->getPic());
  // make it visible in the picture
  mainpane_pic->redrawTextBox();
}


void UserInterface::albumsettingsdlg() 
{
  AlbumOptDlg* dlg = new AlbumOptDlg(myData);
  dlg->showDlg();
  delete dlg;
}

bool UserInterface::IsPermitted(unsigned int uMask)
{
  return(myData->perm->IsPermitted(uMask));
}

void UserInterface::SetPermissions()
{
  if (!IsPermitted(PERM_OverwriteAlbum))
    myMainMenu[MENU_SAVE].deactivate();
  else
    myMainMenu[MENU_SAVE].activate();
}

void UserInterface::Idle()
{  
  if (pDisplayReq != 0)
  {
    TreeItem* item = pDisplayReq;
    bool fAllowFullscreen = fDispReqFullscreen;
    
    lastDisplayedItem = item;
    
    
    // If one is following a link:
    fl_cursor(FL_CURSOR_DEFAULT, FL_FOREGROUND_COLOR, FL_BACKGROUND_COLOR);
    
    if (item->getType() == TreeItem::PIC)
    {
      if (mainpane_group != 0)
      {
        int x;
        int y;
        int w;
        int h;
        
        if (myMainMenu[MENU_FULLSCREEN_PICS].value() && fAllowFullscreen)
        {
          enterFullscreen();
          x = window->x();
          y = window->y();
          w = window->w();
          h = window->h();
        }
        else
        {
          x = mainpane_group->x();
          y = mainpane_group->y();
          w = mainpane_group->w();
          h = mainpane_group->h();
        }
        
        tile->remove(mainpane_group);
        Fl::delete_widget(mainpane_group);
        mainpane_group = (PicGroup*)0;
        browser->connectItemChange(mainpane_group);
        
        if (mainpane_pic == 0)
          mainpane_pic = new PicBox(x, y, w, h);
        else
          mainpane_pic->resize(x, y, w, h);
        mainpane_pic->labelsize(opts->nFontSize);
        mainpane_pic->callback((Fl_Callback*)cb_picbox);
        mainpane_pic->boxcallback((Fl_Callback*)cb_textbox, (void*)this);
        mainpane_pic->align(FL_ALIGN_CENTER);
        tile->add(mainpane_pic);
      }
      
      // load and display pic
      blockingTasks.Pause(true);
      if (mainpane_pic->setPic((TreePicBase*)item,
                               myData->getBasepath(),
                               opts))
      {
        blockingTasks.Pause(false);
        // todo: Fehler absetzen?
        pDisplayReq = 0;
        return;
      }
      blockingTasks.Pause(false);
    }
    
    if (item->getType() == TreeItem::GROUP)
    {
      lastDisplayedGroup = (TreeGroupBase*)item;
      browser->itemSelected = item;
      
      if (mainpane_group == 0)
      {
        int x;
        int y;
        int w;
        int h;
        
        if (windowed != 0)
        {
          if (myMainMenu[MENU_FULLSCREEN].value() == 0)
            window->fullscreen_off(windowed->win_x, windowed->win_y, windowed->win_w, windowed->win_h);
          menuBar->show();
          browser->show();
          x = windowed->pane_x;
          y = windowed->pane_y;
          w = windowed->pane_w;
          h = windowed->pane_h;
          delete windowed;
          windowed = 0;
        }
        else
        {
          x = mainpane_pic->x();
          y = mainpane_pic->y();
          w = mainpane_pic->w();
          h = mainpane_pic->h();
        }
        
        tile->remove(mainpane_pic);
        
        mainpane_group = new PicGroup(x, y, w, h, "");
        mainpane_group->labelsize(opts->nFontSize);
        mainpane_group->setData(myData);
        mainpane_group->setGUI(this);
        mainpane_group->user_data((void*)this);
        mainpane_group->end();
        browser->connectItemChange(mainpane_group);
        mainpane_group->connectItemChange(browser);
        
        tile->add(mainpane_group);
      }
      
      TreeGroupBase*   group = (TreeGroupBase*)item;
      TreeItem*        p;
      TreePicBase*     pic;
      int              nX    = mainpane_group->x() + 5;
      int              nY    = mainpane_group->y() + 5;
      int              nSize = opts->uThumbSize;
      TGroupViewData   gvd;
      std::string      tooltipString;
      std::string      groupdescr = group->getText();
      
      if (groupdescr.length())
        groupdescr = "<p>" + groupdescr + "</p>";
      
      {
        JWWDateTime min, max;
        
        min.nTime = 0;
        min.nDate = 0;
        max = min;
        
        int nCnt = group->getMinMaxDateTime(min, max, false);
        std::string minstr = min.getDateTimeAsString();
        std::string maxstr = max.getDateTimeAsString();
        
        groupdescr += "<p>" + itoStr(nCnt, ' ', 1) + " Pictures</p>";
        
        if (minstr.compare(maxstr) == 0)
        {
          if (maxstr.length())
            groupdescr += "<p>" + maxstr + "</p>";
        }
        else
          groupdescr += "<p>" + maxstr + " to " + minstr + "</p>";
      }
      
      if (group->getCommentHtml().length())
      {
        groupdescr += group->getCommentHtml();
      }
      
      // --- clear group display --------------------------------------
      mainpane_group->clear();
      
      while (groupviewdata.size())
      {
        gvd = groupviewdata[groupviewdata.size()-1];
        groupviewdata.pop_back();
        
        if (gvd.thumbnail != (Fl_Shared_Image*)0)
        {
          gvd.thumbnail->uncache();
          gvd.thumbnail->release();
        }
        free(gvd.tooltip);
      }
      
      // --- make sure thumbnails exist -------------------------------
      if (opts->fCreateThumbs && IsPermitted(PERM_CreateThumbs))
      {
        WorkThumbCreate work;
        
        p = group->itemFirstChild;
        while (p != (TreeItem*)0)
        {
          if (p->getType() == TreeItem::PIC)
          {
            TreePicBase* pic    = (TreePicBase*)p;
            std::string  ref    = myData->getPicPath(pic, 0);
            std::string  result = myData->getPicPath(pic, opts->uThumbSize);
            
            if (FileTools::isFileUpToDate(ref, result) != 1)
            {
              work.filelist.push_back(ref);
              work.filelist.push_back(result);
            }
          }
          p = p->itemNext;
        }
        
        // Create thumbs
        if (work.filelist.size())
        {
          work.nMaxW  = opts->uThumbSize;
          work.nMaxH  = opts->uThumbSize;
          work.flQual = opts->pool->getDouble("thumbnail.quality");
          work.nType  = opts->pool->getInt("thumbnail.mode");
          
          WorkerData   wd;
          wd.init(work.filelist.size()/2, (void*)&work);
          
          int       nRetCode;
          
          if (SimpleThread::Create(createThumbs, (void*)&wd))
          {
            std::cerr << "error creating thread\n";
            std::cerr.flush();
            nRetCode = -1;
          }
          else
          {
            blockingTasks.Pause(true);
            WorkerDlg* wdlg = new WorkerDlg(&wd);
            wdlg->showDlg();
            while (wd.nRun) { Fl::wait(0.2); };
            nRetCode = wd.nRetCode;
            delete wdlg;
            blockingTasks.Pause(false);
          }
          if (nRetCode)
          {
            fl_alert("Error while creating thumbnails (see console)");
          }
        }
      }
      
      // --- display --------------------------------------------------
      setWaitCursor();
      p = group->itemFirstChild;
      while (p != (TreeItem*)0)
      {
        if (p->getType() == TreeItem::PIC && !p->isHidden())
        {
          PicButton* button;
          
          pic = (TreePicBase*)p;
          
          button = new PicButton(nX, nY, nSize+6, nSize+6);
          button->align(FL_ALIGN_CENTER);
          {
            tooltipString = pic->getDescr(4, true);
            gvd.tooltip = (char*)malloc(tooltipString.length()+1);
            strcpy(gvd.tooltip, tooltipString.c_str());
            button->tooltip(gvd.tooltip);
          }
          button->item(pic);
          button->callback( (Fl_Callback*)cb_showItem, (void*)this );
          gvd.thumbnail = Fl_Shared_Image::get(myData->getPicPath(pic, nSize).c_str());
          if (gvd.thumbnail != (Fl_Shared_Image*)0)
          {
            button->image  (gvd.thumbnail);
            button->deimage(gvd.thumbnail);
          }
          else
          {
            button->setLabel(FileTools::name(pic->getText()));
          }
          mainpane_group->add(button);
          
          gvd.pic = pic;
          
          groupviewdata.push_back(gvd);
          
          nY += nSize + 6 + 5;
        }
        
        p = p->itemNext;
      }
      mainpane_group->setDescr(groupdescr);
      mainpane_group->resize(mainpane_group->x(), mainpane_group->y(), mainpane_group->w(), mainpane_group->h());
      mainpane_group->redraw();
      fl_cursor(FL_CURSOR_DEFAULT);
    }
    
    selectCurrentItem();
    
    pDisplayReq = 0;
  }
  Fl::remove_idle(cb_idle, (void*)(this));      
}

void UserInterface::AfterCreation()
{
  updateRecentFilesMenu();
  SetPermissions();
}

int UserInterface::loadDir(const char* dirname, TreeGroupBase* pGroup, TreePicData::eDateTimeSource dtmethod)
{
  if (pGroup == 0)
    pGroup = myData->root;
  setWaitCursor();
  myData->addPicsFromDir(dirname, pGroup, dtmethod);
  fl_cursor(FL_CURSOR_DEFAULT);

  return(0);
}
